/****************************************************************************
 *   st_segmentADT.cc : 
 *   		implementation of segment ADT.
 *
 *   Change Log :
 *		  Dec-24-1996 Initial version,  Liujian Qian
 *
 *   $Id: gadt_segment.cpp,v 1.8 1999/02/16 22:14:47 qian Exp $
 ***************************************************************************/
#ifdef _WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include "gfc.h"

int  _GISegment::cnt;

GSegment::GSegment (const GSegment& a)
{
    x1 = a.x1;     y1 = a.y1;
    x2 = a.x2;     y2 = a.y2;
}

GSegment::GSegment (const GPoint& p1, const GPoint& p2)
{
    set(p1, p2);
}

GSegment::GSegment(int _x1, int _y1, int _x2, int _y2)
{
    set(_x1, _y1, _x2, _y2);
}

uint32  
GSegment::getPSize() const
{
    return (GADT::getPSize()  +   sizeof(Coord)*4 );
}


//
// user format: "x1 y1 x2 y2"
//
Result 
GSegment::input(const char* in)
{
    GAssert(in);
    istrstream is(in, strlen(in));
    is>>x1>>y1>>x2>>y2;

    return GOk;
}


Result 
GSegment::output(char*& out)
{
    bp.recycle();
    char* p = (char*)bp.ptr();
    sprintf(p, " %f %f %f %f ", x1, y1, x2, y2);
    out = p;
    
    return GOk;
}



Result 
GSegment::pack(DataPipe& p) 
{
    
    GADT::pack(p);    
    if(GEndian::big) {
	Coord tmp = x1;
	flipDouble(&tmp);
	p.append(CONPTR(&tmp), sizeof(tmp));
	tmp = y1;
	flipDouble(&tmp);
	p.append(CONPTR(&tmp), sizeof(tmp));
	tmp = x2;
	flipDouble(&tmp);
	p.append(CONPTR(&tmp), sizeof(tmp));
	tmp = y2;
	flipDouble(&tmp);
	p.append(CONPTR(&tmp), sizeof(tmp));
    } else {
	p.append(CONPTR(&x1), sizeof(Coord));
	p.append(CONPTR(&y1), sizeof(Coord));
	p.append(CONPTR(&x2), sizeof(Coord));	
	p.append(CONPTR(&y2), sizeof(Coord));
    }
    return GOk;
}


Result 
GSegment::unpack(DataPipe& p)
{
    GADT::unpack(p);

    GAssert(!p.eop());
    p.fetch((char*)&x1, sizeof(Coord));
    p.fetch((char*)&y1, sizeof(Coord));
    p.fetch((char*)&x2, sizeof(Coord));
    p.fetch((char*)&y2, sizeof(Coord));
    if(GEndian::big) {
	flipDouble(&x1);
	flipDouble(&y1);
	flipDouble(&x2);
	flipDouble(&y2);
    }
    return GOk;
}


int
GSegment::cmp	   (const GADT& a) const
{
    MATCHTYPE(a);
    return UNDEFINED;
}
    

Result
GSegment::scale      (double xf, double yf, double )
{
    x1 *= xf, x2 *= xf;
    y1 *= yf, y2 *= yf;
    return GOk;
}


Result
GSegment::translate      (double xf, double yf, double )
{
    x1 += xf, x2 += xf;
    y1 += yf, y2 += yf;
    return GOk;
}

Result
GSegment::rotate	    (double , bool ) 
{
    return GOk;
}

Result
GSegment::transform  (const GMatrix& ) 
{
    return GOk;
}


Result 
GSegment::mbr(GRect& box) const
{
    if(x1<=x2) 
	box.setXL(x1), box.setXH(x2);
    else 
	box.setXL(x2), box.setXH(x1);
    
    if(y1<=y2)
	box.setYL(y1), box.setYH(y2);
    else
	box.setYL(y2), box.setYH(y1);
    
    return GOk;
}

//
// some handy member functions for setting data members.
//
void    
GSegment::set(Coord _x1, Coord _y1, Coord _x2, Coord _y2) 
{ 
    x1=_x1,y1=_y1,x2=_x2,y2=_y2;
}

void    
GSegment::set(const GPoint&	pt1, const GPoint& pt2)
{
    x1 = pt1.getX();
    y1 = pt1.getY();
    x2 = pt2.getX();
    y2 = pt2.getY();
}


//
// calculate the lenght of the line segment
//
double  
GSegment::length () const
{
    return distPnts(x1,y1, x2,y2);
}


/*
 *-------------------------------------------------------------------------
 * GSegment :: distance (GPoint) --
 *	calculate the vertical distance from the point to the segment.
 *	uses the more advanced distance calculating method below.     
 * Results:
 *	Returns the vertical distance. 
 *	
 * Side effects:
 * 	N/A 
 *-------------------------------------------------------------------------
 */
double  
GSegment::distance (const GPoint& c) const
{
    int dummy;
    
    return distance(c, dummy);
}

/*
 *-------------------------------------------------------------------------
 * GSegment :: distance (GPoint, flag) --
 *	calculate the vertical distance from the point to the segment.
 *	also indicates whether the point' projection falls on the segmetn.
 *
 * Results:
 *	Returns the vertical distance. 
 *	the 'flag' indicates whether the point's perpendicular projection
 *	is on the segment (1) or not (0).                                
 *	
 * Side effects:
 * 	N/A 
 *-------------------------------------------------------------------------
 */
double  
GSegment::distance (const GPoint& c, int& flag) const
{
    double len = length();
    double r,s;
    Coord  xc = c.getX(), yc = c.getY();
    double xI, yI; //point of perpendicular projection of c on the seg.
    
    
    r = ( (y1-yc)*(y1-y2) - (x1-xc)*(x2-x1) ) / (len * len);
    s = ( (y1-yc)*(x2-x1) - (x1-xc)*(y2-y1) ) / (len * len);

    //note: 
    //	   if r<0	I is on backward extension of the seg
    //	   if r>1	I is on ahead extension of the seg
    //     if 0<=r<=1   I is on the seg
    //	
    //	   if s<0	C is left of the seg
    //	   if s>0	C is right of the seg
    //	   if s=0	C is on the seg

    xI = x1 + r*(x2-x1)  ;
    yI = y1 + r*(y2-y1)  ;

    flag = (r>=0 && r<=1)? 1 : 0;
    
    return distPnts(xc, yc, (Coord) xI, (Coord) yI);
}



int	
GSegment::intersects (const GSpatial& a) const
{
    GType at=a.type();
    
    if(at==_Segment) return _intersects((const GSegment&)(a));
    else if(at==_Rect) return _intersects((const GRect&)(a));
    else if(at==_Circle) return _intersects((const GCircle&)(a));
    else if(at==_Arc) return _intersects((const GArc&)(a));
    else if(at==_Polygon) return _intersects((const GPolygon&)(a));
    else msgWarn("Invalid type of operand for Segment.intersects() !");
    
    return UNDEFINED;
}


Result
GSegment::bufferzone(double size,  Coord* xcoords, Coord* ycoords) const
{
    GPoint nv;
    double xoff, yoff, factor;

    //first set the perpendicular vector of this segment
    if(y2>y1)  	nv.x = y1-y2;
    else if(y2<y1)  nv.x = y2-y1;
    else		nv.x = 0;

    if(x2>x1) {
	if(y2>y1)   nv.y = x2-x1;
	else 	nv.y = x1-x2;
    } else if(x2<x1) {
	if(y2>y1)	nv.y = x2-x1;
	else	nv.y = x1-x2;
    } else 		nv.y = 0;
    
    factor = size/nv.distance();	//!!! what if nv.dist=0?
    xoff = nv.x * factor;
    yoff = nv.y * factor;
    //first vertice is the one perpendicular to (x1, y1) and to
    //the left side of this segment.
    xcoords[0] = x1 + xoff;
    ycoords[0] = y1 + yoff;
    //second vertice is the one perpendicular to (x1, y1) and to
    //the right side of this segment.
    xcoords[1] = x1 - xoff;
    ycoords[1] = y1 - yoff;
    //third vertice is the one perpendicular to (x2, y2) and to
    //the right side of this segment.
    xcoords[2] = x2 - xoff;
    ycoords[2] = y2 - yoff;
    //forth vertice is the one perpendicular to (x2,y2) and to 
    //the left side of this segment.
    xcoords[3] = x2 + xoff;
    ycoords[3] = y2 + yoff;
    //the last vetice repeats first one to make it fclosed.
    xcoords[4] = xcoords[0];
    ycoords[4] = ycoords[0];

    return GOk;    
}



/*
 *-----------------------------------------------------------------------
 *   linesIntersect()
 *
 *   This function computes whether two line segments,
 *   respectively joining the input points (x1,y1) -- (x2,y2)
 *   and the input points (x3,y3) -- (x4,y4) intersect.
 *   If the lines intersect, the output variables x, y are
 *   set to coordinates of the point of intersection.
 *
 *   Originally written by Mukesh Prasad (In "Graphics Gems")
 *-----------------------------------------------------------------------
 */
#define	DONT_INTERSECT    0
#define	DO_INTERSECT      1
#define COLLINEAR         2

int linesIntersect ( Coord x1, Coord y1,   // First line segment 
		     Coord x2, Coord y2,
		     Coord x3, Coord y3,   // Second line segment 
		     Coord x4, Coord y4,
		     Coord& x,      // Output value: point of intersection 
		     Coord& y         
		    )
{
    Coord a1, a2, b1, b2, c1, c2; // Coefficients of line eqns.
    Coord r1, r2, r3, r4;         // 'Sign' values 
    Coord denom, offset, num;     // Intermediate values

    // Compute a1, b1, c1, where line joining points 1 and 2
    // is "a1 x  +  b1 y  +  c1  =  0".
     
    a1 = y2 - y1;
    b1 = x1 - x2;
    c1 = x2 * y1 - x1 * y2;

    // Compute r3 and r4.

    r3 = a1 * x3 + b1 * y3 + c1;
    r4 = a1 * x4 + b1 * y4 + c1;

    // Check signs of r3 and r4.  If both point 3 and point 4 lie on
    // same side of line 1, the line segments do not intersect.
    //

    if ( r3 != 0 &&
         r4 != 0 &&
         SAME_SIGNS( r3, r4 ))
        return ( DONT_INTERSECT );

    // Compute a2, b2, c2 

    a2 = y4 - y3;
    b2 = x3 - x4;
    c2 = x4 * y3 - x3 * y4;

    // Compute r1 and r2 

    r1 = a2 * x1 + b2 * y1 + c2;
    r2 = a2 * x2 + b2 * y2 + c2;

    // Check signs of r1 and r2.  If both point 1 and point 2 lie
    // on same side of second line segment, the line segments do
    // not intersect.

    if ( r1 != 0 &&
         r2 != 0 &&
         SAME_SIGNS( r1, r2 ))
        return ( DONT_INTERSECT );

    // Line segments intersect: compute intersection point. 

    denom = a1 * b2 - a2 * b1;
    if ( denom == 0 )
        return ( COLLINEAR );
    offset = denom < 0 ? - denom / 2 : denom / 2;

    // The denom/2 is to get rounding instead of truncating.  It
    // is added or subtracted to the numerator, depending upon the
    // sign of the numerator.
    
    num = b1 * c2 - b2 * c1;
    //this seems to causing an slight error in the result
    //x = ( num < 0 ? num - offset : num + offset ) / denom;
    x = num/denom;
    
    num = a2 * c1 - a1 * c2;
    //y = ( num < 0 ? num - offset : num + offset ) / denom;
    y = num/denom;
    
    return ( DO_INTERSECT );
}

int
GSegment::_intersects (const GSegment& seg, GPoint* secp) const
{
    Coord sec_x, sec_y;
    
    if( linesIntersect(x1,y1, x2,y2,  
		       seg.x1, seg.y1, seg.x2, seg.y2,
		       sec_x, sec_y) == DO_INTERSECT ) {
	if(secp) { //pass intersecting point out.
	    secp->setX(sec_x);
	    secp->setY(sec_y);
	}
	return 1;
    }
    else return 0;
}



//
// check if this segment intersects a rectangle.
//
int
GSegment::_intersects (const GRect&  r) const
{
    GSegment edge;
    Coord xlow = r.getXL();    Coord ylow = r.getYL();
    Coord xhi  = r.getXH();    Coord yhi  = r.getYH();
    
    edge.set(xlow, ylow, xlow, yhi);
    if(intersects(edge)) return 1;
    
    edge.set(xlow, yhi, xhi, yhi);
    if(intersects(edge)) return 1;
    
    edge.set(xhi,yhi, xhi, ylow);
    if(intersects(edge)) return 1;
    
    edge.set(xhi, ylow, xlow, ylow);
    if(intersects(edge)) return 1;

    return 0;
}


//
// check if a segment intersects a circle; just see 
//
int
GSegment::_intersects (const GCircle& c) const
{
    GPoint pnt;
    
    pnt.set (x1, y1);

    if( pnt.isInside(c) ) {
	pnt.set (x2, y2);
	if(pnt.isInside(c) ) return 0;
	else return 1;
    }
    else {
	pnt.set (x2, y2) ;
	if( pnt.isInside(c) ) return 1;
	else {     // both ends outside the circle. 
	    int proj_on_seg;
	    
	    pnt.set(c.getCX(), c.getCY());
	    double dist = distance(pnt, proj_on_seg);
	    if(dist <= c.getR() && proj_on_seg ) return 1;
	    else return 0;
	}
    }
}


//
// check if this segment intersects a polyline.
// call polyline's intersects() method to do the hard work.
//
int    
GSegment::_intersects (const GArc& arc) const
{
    return arc.intersects(*this);
}


//
// check if this segment intersects a polygon.
// call polygon's intersects() method to do the hard work.
//
int
GSegment::_intersects (const GPolygon& pgn) const
{
    return pgn.intersects (*this);
}



//
// check if this segment is inside a rectangle.
// simply use the bounding rectangle and call
// covers() method of incoming rectangle.     
//
int    
GSegment::isInside     (const GRect& r) const
{
    return r.covers(*this);
}

//
// check if the segment is inside a circle.
// 
int    
GSegment::isInside     (const GCircle& other) const
{
    return other.covers(*this);
    
}

ostream& operator << (ostream& s, const GSegment& g)
{
    return s<<"("<<g.x1<<", "<<g.y1<<"| "<<g.x2<<", "<<g.y2<<")";
}

