/****************************************************************************
 *   datapipe.cc : 
 *   		implementation of DataPipe class.
 *
 *   Change Log :
 *		 July-12 1997 Initial version,  Liujian Qian
 *
 *   $Id: gdatapipe.cpp,v 1.5 1999/02/18 05:58:22 qian Exp $
 ***************************************************************************/

#include <string.h>
#include "gdatapipe.h"


//
/// copy ctor
//
DataPipe::DataPipe(const DataPipe& a)
{
    data_len = _capacity = cur = 0;
    udp = 0;
    append(a.ptr(), a.size());
    cur = 0;   //be nice to reset cursor to the begining
}

DataPipe::DataPipe(int len)
{
    data_len = _capacity = cur = 0;
    udp = 0;
    extendTo(len);
    cur = 0;
}

//
// this constructor is not for attaching; it actulay copy
// the data pointed by ptr into newly allocated pipe space.
DataPipe::DataPipe(const char* ptr, int length)
{
    data_len = _capacity = cur = attached = 0;
    udp = 0;
    append(ptr, length);
    cur = 0; //be nice to reset the cusor to begining
}

DataPipe::~DataPipe()
{
    if(!attached && udp) delete[] udp;
}

DataPipe&
DataPipe::operator= (const DataPipe& a)
{
    if(_capacity < a.size())
	if( extend(a.size()-_capacity) != GOk) {
	    msgWarn("DataPipe: out of memory!");
	    return *this;
	} 
    recycle();
    append(a.ptr(), a.size());
    cur = 0;
    return *this;
}

/**
 ** append a chunk of data at the end of the data; extend the piep
 ** if out of range ; cursor is set to the new end of data
 */
Result 
DataPipe::append(const char* ptr, int sz)
{
    if(!ptr || sz<=0) {
	return GOk;
    }
    
    if(data_len+sz >_capacity)
	if(extend(data_len+sz-_capacity) != GOk) msgFail("DataPipe: out of memory!");
    memcpy(udp+data_len, ptr, sz);
    data_len += sz;
    cur = data_len;
    return GOk;
}

//
/// insert a chunk of data at the beginning 
/// cursor is set to new end of data
//
Result 
DataPipe::insert(const char* ptr, int sz)
{ 
    if(!ptr || sz<=0) {
	return GOk;
    }

    if(data_len+sz>_capacity)
	if(extend(sz) != GOk) msgFail("DataPipe: out of memory!");
    memmove(udp+sz, udp, data_len);
    memcpy(udp, ptr, sz);
    data_len += sz;
    cur = data_len;
    return GOk;
}

//
/// copy 'length' of data starting at cursor into 'dest'.
//  the cursor is advanced automaticaly.
//
Result 
DataPipe::fetch(char* dest, int length)
{
    if(!dest || length<=0) {
	return GOk;
    }

    if( cur<0 || cur>=data_len || cur+length>data_len )
	msgFail("pipe: attempt to fetch invalid range!");
    
    memcpy(dest, udp+cur, length);
    cur += length;
    return GOk;
}

//
// 
Result 
DataPipe::attach(void* p, int space)
{
    if(!p) msgFail("can't attach to void pointer!");
    if(attached) detach(); //detach any prevoulsy attached pointer

    attached = 1;
    if(udp) delete [] udp;
    udp = (char*)p;
    _capacity = data_len = space;
    //data_len = 0;
    cur = 0;
    return GOk;
}

//
//
Result
DataPipe::detach(void)
{
    if(!attached) msgFail("not previously attached!");
    
    udp = 0;
    _capacity = cur = data_len = 0;
    attached = 0;
    return GOk;
}


/**
 ** move the cursor to the specified position
 */
Result 
DataPipe::seek(int off, int dir) 
{
    if(dir == BEGIN) cur = off;
    else cur += off;

    if(cur >= _capacity) return GBound;     //out of boundary
    //else if(cur >= data_len) msgWarn("DataPipe: accessing noninitialized portion");
    return GOk;
}


/// increase the size the pipe's array; default is to increase delta bytes each time
//
Result 
DataPipe::extend(int sz)
{
    if(attached) msgFail("can't extend an attached pointer!");
    
    if(sz<=0) return GOk; //can't shrink

    int incr = (sz<delta)?delta:sz;
    char* tmp = new char[_capacity + incr];

    if(!tmp) return GFail;

    _capacity +=  incr;
    
    if(udp) {
	memcpy(tmp, udp, data_len); //save old data
	delete[] udp; 
    }
    udp = tmp;

    return GOk;
}

//
Result
DataPipe::extendTo(int new_cap) 
{
    if(attached) msgFail("can't extendTo attached pointer!");
    
    int incr = (new_cap-_capacity);
    return extend(incr);
}


ostream& operator<<(ostream& s, const DataPipe& p)
{
    s<<"(capacity="<<p._capacity<<", data="<<p.size()<<", cursor="<<p.cur<<")"<<endl;
    return s;
}

///
// some handy pipe operator functions

DataPipe& operator<< (DataPipe& s, const char& c)
{
    s.append( CONPTR(&c), sizeof(char));
    return s;
}

DataPipe& operator<< (DataPipe& s, const unsigned char&  c)
{
    s.append( CONPTR(&c), sizeof(unsigned char));
    return s;
}

DataPipe& operator<< (DataPipe& s, const int& i)
{
    s.append(CONPTR(&i), sizeof(int));
    return s;
}

DataPipe& operator<< (DataPipe& s, const short& t)
{
    s.append(CONPTR(&t), sizeof(short));
    return s;
}

DataPipe& operator<< (DataPipe& s, const float& f)
{
    s.append(CONPTR(&f), sizeof(float));
    return s;
}

DataPipe& operator<< (DataPipe& s, const double& d)
{
    s.append(CONPTR(&d), sizeof(double));
    return s;
}

DataPipe& operator<< (DataPipe& s, const long& i)
{
    s.append(CONPTR(&i), sizeof(long));
    return s;
}

DataPipe& operator<< (DataPipe& s, const unsigned int& i)
{
    s.append(CONPTR(&i), sizeof(unsigned int));
    return s;
}

DataPipe& operator<< (DataPipe& s, const unsigned short& t)
{
    s.append(CONPTR(&t), sizeof(unsigned short));
    return s;
}

DataPipe& operator<< (DataPipe& s, const unsigned long& i)
{
    s.append(CONPTR(&i), sizeof(unsigned long));
    return s;
}

#ifndef _WIN32
DataPipe& operator<< (DataPipe& s, const int64& i)
{
        s.append(CONPTR(&i), sizeof(int64));
        return s;
}
#endif

DataPipe& operator<< (DataPipe& s, const char* string)
{
    register int len=0;
    register const char* p=string;
    while(*p++) len++;
    s.append(string, len+1);	//tailing '\0' is also piped.
    return s;
}
    

///
//

DataPipe& operator>> (DataPipe& s, char& c)
{
    s.fetch( (char*)(&c), sizeof(char));
    return s;
}

DataPipe& operator>> (DataPipe& s, unsigned char&  c)
{
    s.fetch( (char*)(&c), sizeof(unsigned char));
    return s;
}

DataPipe& operator>> (DataPipe& s, int& i)
{
    s.fetch((char*)(&i), sizeof(int));
    return s;
}

DataPipe& operator>> (DataPipe& s, short& t)
{
    s.fetch((char*)(&t), sizeof(short));
    return s;
}

DataPipe& operator>> (DataPipe& s, float& f)
{
    s.fetch((char*)(&f), sizeof(float));
    return s;
}

DataPipe& operator>> (DataPipe& s, double& d)
{
    s.fetch((char*)(&d), sizeof(double));
    return s;
}

DataPipe& operator>> (DataPipe& s, long& i)
{
    s.fetch((char*)(&i), sizeof(long));
    return s;
}

DataPipe& operator>> (DataPipe& s, unsigned int& i)
{
    s.fetch((char*)(&i), sizeof(unsigned int));
    return s;
}

DataPipe& operator>> (DataPipe& s, unsigned short& t)
{
    s.fetch((char*)(&t), sizeof(unsigned short));
    return s;
}

DataPipe& operator>> (DataPipe& s, unsigned long& i)
{
    s.fetch((char*)(&i), sizeof(unsigned long));
    return s;
}

#ifndef _WIN32
DataPipe& operator>> (DataPipe& s, int64& i)
{
        s.fetch((char*)(&i), sizeof(int64));
        return s;        
}        
#endif

DataPipe& operator>> (DataPipe& s, char* string)
{
    register int len=0;
    register char* p=s.cursor();
    while(*p++) len++;
    s.fetch(string, len+1);	//tailing '\0' is also fetched.
    return s;
}
    

