/****************************************************************************
 *   gmatrix.cpp : 
 *   		  implementation of GMatrix class.
 *   Change Log :
 *		  Jan 31 1998 Initial version,  Liujian Qian
 *
 *   $Id: gmatrix.cpp,v 1.3 1998/12/10 00:11:39 qian Exp $
 ***************************************************************************/
#include <math.h>
#include "gmatrix.h"
#include "gdatapipe.h"

GMatrix::GMatrix()
{
    init();
}

GMatrix::GMatrix(double m11, double m12, double m21, double m22,
		 double _dx,  double _dy )
{
    set(m11, m12, m21, m22, _dx, _dy);
}

Result
GMatrix::init() 
{
    a11 = a22 = 1.0;
    a12 = a21 = dx = dy = 0.0;
    return GOk;
}

void
GMatrix::set(double m11, double m12, double m21, double m22,
	     double _dx,  double _dy )
{
    a11 = m11, a12 = m12, a21 = m21, a22 = m22;
    dx = _dx, dy = _dy;
}


void
GMatrix::xform(int x, int y, int& new_x, int& new_y) const
{
    new_x = (int)(a11*x + a21*y + dx);
    new_y = (int)(a12*x + a22*y + dy);
}

void
GMatrix::xform(double x, double y, double& new_x, double& new_y) const
{
    new_x = (a11*x + a21*y + dx);
    new_y = (a12*x + a22*y + dy);
}

void
GMatrix::xform(int n_pnt, double* points, double* new_points)
{
    double x,y,nx,ny;

    for(int i=0; i<n_pnt; i++) {
	x=points[i*2], y=points[i*2+1];
	xform(x,y,nx,ny);
	new_points[i*2] = nx, new_points[i*2+1] = ny;
    }
}

GMatrix&
GMatrix::translate(double _dx, double _dy)
{
    GMatrix	res (1.0F, 0.0F, 0.0F, 1.0F, _dx, _dy);
    return mult(res);
}

GMatrix&
GMatrix::scale(double sx, double sy)
{
    GMatrix res( sx, 0.0F, 0.0F, sy, 0.0F, 0.0F );
    return mult( res );
}

GMatrix&	
GMatrix::shear(double sh, double sv)
{
    GMatrix res( 1.0F, sv, sh, 1.0F, 0.0F, 0.0F );
    return mult( res );
}

GMatrix&	
GMatrix::rotate( double a)
{
    double b = DEG_TO_RAD*a;			// convert to radians
    double sina = sin(b);
    double cosa = cos(b);
    GMatrix res( cosa, sina, -sina, cosa, 0.0F, 0.0F );
    return mult( res );
}
    

GMatrix	
GMatrix::invert(bool& invertable) const
{
    double det = a11*a22 - a12*a21;
    if ( fabs(det) < 0.000001 ) {		// very close to zero
	invertable = FALSE;			// singular matrix
	GMatrix defaultMatrix;
	return defaultMatrix;
    }
    else {					// invertible matrix
	invertable = TRUE;
	double dinv = 1.0/det;
	GMatrix imatrix( (a22*dinv),   (-a12*dinv),
			 (-a21*dinv),  (a11*dinv),
			 ((a21*dy - a22*dx)*dinv),
			 ((a12*dx - a11*dy)*dinv) );
	return imatrix;
    }
}
    

GMatrix&	
GMatrix::operator*= ( const GMatrix& m)
{
    set( a11*m.a11 + a12*m.a21,  a11*m.a12 + a12*m.a22,
	 a21*m.a11 + a22*m.a21,  a21*m.a12 + a22*m.a22,
	 dx*m.a11 + dy*m.a21 + m.dx,
	 dx*m.a12 + dy*m.a22 + m.dy );
    return *this;
}
    

Result	
GMatrix::pack	(DataPipe& p) const 
{
    p<<a11<<a12<<a21<<a22<<dx<<dy;
    return GOk;
}

Result
GMatrix::unpack  (DataPipe& p) 
{
    p>>a11>>a12>>a21>>a22>>dx>>dy;
    return GOk;
}


GMatrix&	
GMatrix::mult(const GMatrix& m)
{
    set( m.a11*a11 + m.a12*a21,  m.a11*a12 + m.a12*a22,
	 m.a21*a11 + m.a22*a21,  m.a21*a12 + m.a22*a22,
	 m.dx*a11 + m.dy*a21 + dx,
	 m.dx*a12 + m.dy*a22 + dy );
    return *this;
}

GMatrix		operator* (const GMatrix& a, const GMatrix& b)
{
    GMatrix res = a;
    res *= b;
    return res;
}


 




