/****************************************************************************
 *   gproj.cpp : 
 *   		 implementation of basic geographic referencing systems.	
 *   Credit: 
 *	       The map projections in this file were implemented originaly by
 *	       EROS Data Center in the package GCTPC.  
 *	       (ftp://edcftp.cr.usgs.gov:pub/software/gctpc)
 *             This file and gproj.cpp merely provide a c++ wrap up of the
 *             projections. Each projection is implemented in an individual class.
 *
 *   Change Log :
 *		 May 25, 1998 Initial version,  Liujian Qian
 *
 *   $Id: gproj.cpp,v 1.19 1998/12/10 00:11:42 qian Exp $
 ***************************************************************************/
#include <math.h>
#include "gproj.h"

GEllipsoid clarke66 = {6378206.4, 6356583.8};
GEllipsoid new_international = {6378157.5, 6356772.2};


//////////////////////////////////////////////////////////////////////
// class GMapProj

GMapProj::GMapProj()
{
    proj_id = Geographic;	//default is Geographic (lat/lon) reference
}

uint32  GMapProj::getPSize() const 
{
    return GObject::getPSize() + 
	sizeof(GEllipsoid) + sizeof(double)*4 + 
	sizeof(GProjCode);
}

ostream&    operator<<(ostream& s, const GMapProj& )
{
    return s<<"Projection: geographic\n";
}

Result
GMapProj::input( const char* )
{
    //istrstream is(in, strlen(in));

    //do nothing here
    //is>>x>>y;
    
    return GOk;
}

Result
GMapProj::output(char*& )
{
    return GOk;
}

Result
GMapProj::pack(DataPipe& p) 
{
    GObject::pack(p);
    
    p.append(CONPTR(&ellip), sizeof(GEllipsoid));  //ellipsoid
    p.append(CONPTR(&lon_center), sizeof(double));  //central longi
    p.append(CONPTR(&false_easting), sizeof(double));
    p.append(CONPTR(&false_northing), sizeof(double));
    p.append(CONPTR(&proj_id), sizeof(GProjCode));

    return GOk;
}

Result 
GMapProj::unpack(DataPipe& p)
{
    GObject::unpack(p);
    GAssert(!p.eop());
    
    p.fetch((char*)&ellip, sizeof(GEllipsoid));  //ellipsoid
    p.fetch((char*)&lon_center, sizeof(double));  //central longi
    p.fetch((char*)&false_easting, sizeof(double));
    p.fetch((char*)&false_northing, sizeof(double));
    p.fetch((char*)&proj_id, sizeof(GProjCode));
    
    return GOk;
}

double GMapProj::adjustLon(double x)
{
    long count = 0;
    for(;;)
	{
	    if (fabs(x)<=PI)
		break;
	    else if (((long) fabs(x / PI)) < 2)
		    x = x-(GSGN(x) *TWO_PI);
	    else if (((long) fabs(x / TWO_PI)) < MAXLONG){
		x = x-(((long)(x / TWO_PI))*TWO_PI);
	    }
	    else if (((long) fabs(x / (MAXLONG * TWO_PI))) < MAXLONG){
		x = x-(((long)(x / (MAXLONG * TWO_PI))) * (TWO_PI * MAXLONG));
	    }
	    else if (((long) fabs(x / (DBLLONG * TWO_PI))) < MAXLONG) {
		x = x-(((long)(x / (DBLLONG * TWO_PI))) * (TWO_PI * DBLLONG));
	    }
	    else
		x = x-(GSGN(x) *TWO_PI);
	    count++;
	    if (count > MAX_VAL)
		break;
	}

    return(x);
}

double GMapProj::asinZ(double x)
{
    if (fabs(x) > 1.0)  {
	if (x > 1.0)
	    x = 1.0;
	else
	    x = -1.0;
    }
    return(asin(x));
}

double GMapProj::e0(double x) 
{
    return(1.0-0.25*x*(1.0+x/16.0*(3.0+1.25*x)));
}

double GMapProj::e1(double x)
{
    return(0.375*x*(1.0+0.25*x*(1.0+0.46875*x)));
}

double GMapProj::e2(double x)
{
    return(0.05859375*x*x*(1.0+0.75*x));
}

double GMapProj::e3(double x)
{
    return(0.011393229*x*x*x);
}

double GMapProj::e4(double x)
{
    double con = 1.0 + x; 
    double com = 1.0 - x;
    return (sqrt((pow(con,con))*(pow(com,com))));
}

double	GMapProj::ms(double eccent, double sinphi, double cosphi)
{
    double con = eccent*sinphi;
    return ((cosphi/(sqrt(1.0-con*con))));
}

double  GMapProj::qs(double eccent, double sinphi, double )
{

   if (eccent > 1.0e-7) {
       double con = eccent * sinphi;
       return (( 1.0- eccent * eccent) * (sinphi /(1.0 - con * con) - (.5/eccent)*
					  log((1.0 - con)/(1.0 + con))));
   } else
       return(2.0 * sinphi);
}

double  GMapProj::ts(double eccent, double phi, double sinphi)
{
    double con = eccent * sinphi;
    double com = .5 * eccent; 
    con = pow(((1.0 - con) / (1.0 + con)),com);
    return (tan(.5 * (HALF_PI - phi))/con);
}    

double  GMapProj::phi1z(double eccent, double qs, Result& res)
{
    double dphi, con, com, sinphi, cosphi;

    res = GOk;
    double phi = asinZ(.5 * qs);
    if (eccent < EPSLN) 
	return(phi);
    double eccnts = eccent * eccent; 
    for (int i = 1; i <= 25; i++) {
        sinphi = sin(phi), cosphi = cos(phi);
        con = eccent * sinphi; 
        com = 1.0 - con * con;
        dphi = .5 * com * com / cosphi * (qs / (1.0 - eccnts) - sinphi / com + 
               .5 / eccent * log ((1.0 - con) / (1.0 + con)));
       phi = phi + dphi;
       if (fabs(dphi) <= 1e-7)
	   return(phi);
    }
    msgWarn("phi1z: Convergence error");
    res = GFail;
    return -1.0;	//error!
}    

double GMapProj::phi2z(double eccent, double ts, Result& res)
{
    double con;
    double dphi;
    double sinpi;

    res = GOk;
    double eccnth = .5 * eccent;
    double phi = HALF_PI - 2 * atan(ts);
    for (int i = 0; i <= 15; i++) {
	sinpi = sin(phi);
	con = eccent * sinpi;
	dphi = HALF_PI - 2 * atan(ts *(pow(((1.0 - con)/(1.0 + con)),eccnth))) - 
	    phi;
	phi += dphi; 
	if (fabs(dphi) <= .0000000001)
	    return(phi);
    }
    msgWarn("phi2z: Convergence error");
    res = GFail;
    return(-1);
}

double  GMapProj::phi3z(double ml,double e0,double e1,
			double e2, double e3, Result& res)
{
    double dphi;

    double phi = ml;
    for (int i = 0; i < 15; i++) {
	dphi = (ml + e1 * sin(2.0 * phi) - e2 * sin(4.0 * phi) + e3 * sin(6.0 * phi))
	    / e0 - phi;
	phi += dphi;
	if (fabs(dphi) <= .0000000001) {
	    res = GOk;
	    return(phi);
	}
    }

    msgWarn("phi3z: Latitude failed to converge after 15 iterations");
    res = GFail;
    return(-1);
}

Result  GMapProj::phi4z(double eccent, double e0, double e1, double e2, double e3,
			double a, double b, double* c, double* phi)
{
    double sinphi;
    double sin2ph;
    double tanphi;
    double ml;
    double mlp;
    double con1;
    double con2;
    double con3;
    double dphi;

      *phi = a;
      for (int i = 1; i <= 15; i++) {
	  sinphi = sin(*phi);
	  tanphi = tan(*phi);
	  *c = tanphi * sqrt (1.0 - eccent * sinphi * sinphi);
	  sin2ph = sin (2.0 * *phi);

	  ml = e0 * *phi - e1 * sin2ph + e2 * sin (4.0 *  *phi) - e3 * 
	      sin (6.0 *  *phi);
	  mlp = e0 - 2.0 * e1 * cos (2.0 *  *phi) + 4.0 * e2 *
              cos (4.0 *  *phi) - 6.0 * e3 * cos (6.0 *  *phi);
	  con1 = 2.0 * ml + *c * (ml * ml + b) - 2.0 * a *  (*c * ml + 1.0);
	  con2 = eccent * sin2ph * (ml * ml + b - 2.0 * a * ml) / (2.0 * *c);
	  con3 = 2.0 * (a - ml) * (*c * mlp - 2.0 / sin2ph) - 2.0 * mlp;
	  dphi = con1 / (con2 + con3);
	  *phi += dphi;
	  if (fabs(dphi) <= .0000000001 )
	      return(GOk);
        }
      msgFail("phi4z: Lattitude failed to converge");
}   

double
GMapProj::distML(double phi, double e0, double e1,
		 double e2, double e3) const
{
    return(e0*phi-e1*sin(2.0*phi)+e2*sin(4.0*phi)-e3*sin(6.0*phi));
}

int
GMapProj::getUTMZone(double lon)
{
    return((int)(((lon + 180.0) / 6.0) + 1.0)); 
}

double  GMapProj::DMS2DD(int d, int m, int s)
{
    return (d*3600.0+m*60+s)/3600.0;
}

void    GMapProj::DD2DMS(double dd, int& d, int& m, int& s)
{
    double sec = dd*3600.0;
    d = (int) (sec/3600.0);
    sec -= d*3600.0;
    m = (int) (sec/60);
    sec -= m*60.0;
    if(sec>0) s = (int) ceil(sec);
    else if(sec<0) s = (int)(floor(sec));
    else s = (int)sec;
}


bool GMapProj::isClarke66() const 
{
    return (ellip.major == clarke66.major && 
	    ellip.minor == clarke66.minor)? 1:0;
}

bool GMapProj::isSphere(const double c) const
{
    double _es = 1-(ellip.minor/ellip.major)*(ellip.minor/ellip.major);
    return (_es<c)? 1:0;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// GProjUTM

GProjUTM::GProjUTM()
{
    proj_id = UTM;
}

GProjUTM::GProjUTM(double r_maj, double r_min, int _zone)
{
    init(r_maj, r_min, _zone);
}

GProjUTM::GProjUTM(const GEllipsoid& e, int _zone)
{
    init(e.major, e.minor, _zone);
}

ostream&    operator<<(ostream& s, const GProjUTM& a)
{
    s<<"UTM: Zone="<<a.zone<<", major="<<a.ellip.major<<", minor=";
    s<<a.ellip.minor<<endl;
    s<<"UTM: false easting="<<a.false_easting<<", false northing=";
    return s<<a.false_northing<<", scale facotr="<<a.scale_factor<<", sphere="<<a.indicator<<endl;
}

uint32 GProjUTM::getPSize() const {
    return GMapProj::getPSize() + sizeof(int);  //zone
}
Result GProjUTM::input     (const char*  ) {
    return GOk;
}
Result GProjUTM::output    (char*& ) {
    return GOk;
}
Result GProjUTM::pack	  (DataPipe& p) {
    GMapProj::pack(p);
    p.append(CONPTR(&zone), sizeof(int));
    return GOk;
}
Result GProjUTM::unpack    (DataPipe& p ) {
    GMapProj::unpack(p);
    GAssert(!p.eop());
    p.fetch((char*)&zone, sizeof(int));
    return GOk;
}

Result
GProjUTM::init(double r_maj, double r_min, int _zone)
{
    proj_id = UTM;

    ellip.major = r_maj, ellip.minor = r_min;
    scale_factor  = 0.9996;  //scale on central meridian
    lat_origin    = 0.0;
    zone = _zone;
    lon_center    = ((6 * GABS(zone)) - 183) * DEG_TO_RAD;
    false_easting = 500000.0;
    false_northing = (_zone < 0) ? 10000000.0 : 0.0;
    double temp = r_min/r_maj;
    es = 1.0 - temp*temp;
    _e = sqrt(es), _e0 = e0(es), _e1 = e1(es), _e2 = e2(es), _e3 = e3(es);
    ml0 = r_maj * distML(_e0,_e1,_e2,_e3,lat_origin);
    _esp = es / (1.0 - es);
    
    indicator = isSphere();

    return GOk;
}

Result
GProjUTM::forward(double lon, double lat, double& x, double& y)
{
    double	delta_lon;
    double 	sin_phi, cos_phi;
    double	al, als, b, c, t, tq;
    double	con, n, ml;

    delta_lon = adjustLon(lon - lon_center);
    //delta_lon = (lon - lon_center);
    
    sin_phi = sin(lat), cos_phi = cos(lat);
    
    if(indicator) {
	b = cos_phi * sin(delta_lon);
	if (( fabs(fabs(b)-1.0)) < .0000000001) {
	    msgFail("UTM Forward: projecting into infinity!");
	} else {
	    x = 0.5 * ellip.major * scale_factor * log((1.0 + b)/(1.0 - b));
	    con = acos(cos_phi * cos(delta_lon)/sqrt(1.0 - b*b));
	    if (lat < 0)
		con = - con;
	    y = ellip.major * scale_factor * (con - lat_origin); 
	    return(GOk);
	}
    }
    
    al  = cos_phi * delta_lon;
    als = al*al;
    c   = _esp * (cos_phi*cos_phi);
    tq  = tan(lat);
    t   = tq*tq;
    con = 1.0 - es * (sin_phi*sin_phi);
    n   = ellip.major / sqrt(con);
    ml  = ellip.major * distML(lat, _e0, _e1, _e2, _e3);

    x   = scale_factor * n * al * (1.0 + als / 6.0 * (1.0 - t + c + als / 20.0 *
	  (5.0 - 18.0 * t + t*t + 72.0 * c - 58.0 * _esp))) + false_easting;

    y   = scale_factor * (ml - ml0 + n * tq * (als * (0.5 + als / 24.0 *
          (5.0 - t + 9.0 * c + 4.0 * (c*c) + als / 30.0 * (61.0 - 58.0 * t
          + t*t + 600.0 * c - 330.0 * _esp))))) + false_northing;

    return(GOk);
}


Result
GProjUTM::inverse(double x, double y, double& lon, double& lat)
{
    double con,phi;	// temporary angles		     
    double delta_phi;	// difference between longitudes
    long i;		// counter variable
    double sin_phi, cos_phi, tan_phi;	// sin cos and tangent values
    double c, cs, t, ts, n, r, d, ds;	// temporary variables
    double f, h, g, temp;		// temporary variables		
    long max_iter = 6;			// maximun number of iterations

    if (indicator != 0)  {
	f = exp(x/(ellip.major * scale_factor));
	g = .5 * (f - 1/f);
	temp = lat_origin + y/(ellip.major * scale_factor);
	h = cos(temp);
	con = sqrt((1.0 - h * h)/(1.0 + g * g));
	lat = asinZ(con);
	if (temp < 0)
	    lat = -lat;
	if ((g == 0) && (h == 0))   {
		lon = lon_center;
		return(GOk);
	} else   {
		lon = adjustLon(atan2(g,h) + lon_center);
		return(GOk);
	}
    }
    x = x - false_easting;
    y = y - false_northing;

    con = (ml0 + y / scale_factor) / ellip.major;
    phi = con;
    for (i=0;;i++)  {
	delta_phi=((con + _e1 * sin(2.0*phi) - _e2 * sin(4.0*phi) + _e3 * sin(6.0*phi))
		   / _e0) - phi;
	/*
	  delta_phi = ((con + e1 * sin(2.0*phi) - e2 * sin(4.0*phi)) / e0) - phi;
	  */
	phi += delta_phi;
	if (fabs(delta_phi) <= EPSLN) break;
	if (i >= max_iter) { 
	    msgFail("UTM Inverse: Latitude failed to converge!"); 
	}
    }
    if (fabs(phi) < HALF_PI)  {
	sin_phi = sin(phi), cos_phi = cos(phi);
	tan_phi = tan(phi);
	c    = _esp * (cos_phi *cos_phi);
	cs   = (c*c);
	t    = (tan_phi*tan_phi);
	ts   = (t*t);
	con  = 1.0 - es * (sin_phi*sin_phi); 
	n    = ellip.major / sqrt(con);
	r    = n * (1.0 - es) / con;
	d    = x / (n * scale_factor);
	ds   = (d*d);
	lat = phi - (n * tan_phi * ds / r) * (0.5 - ds / 24.0 * (5.0 + 3.0 * t + 
	       10.0 * c - 4.0 * cs - 9.0 * _esp - ds / 30.0 * (61.0 + 90.0 * t +
               298.0 * c + 45.0 * ts - 252.0 * _esp - 3.0 * cs)));
        lon = adjustLon(lon_center + (d * (1.0 - ds / 6.0 * (1.0 + 2.0 * t +
               c - ds / 20.0 * (5.0 - 2.0 * c + 28.0 * t - 3.0 * cs + 8.0 * _esp +
               24.0 * ts))) / cos_phi));
    }
    else  {
	lat = HALF_PI * GSGN(y);
	lon = lon_center;
    }
    return(GOk);
}



//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// Lambert Conformal Conic

GProjLambert::GProjLambert()
{
    proj_id = Lambert;    
}

ostream&    operator<<(ostream& s, const GProjLambert& a)
{
    s<<"Lambert(CC): major="<<a.ellip.major<<", minor=";
    s<<a.ellip.minor<<", std_par1="<<a.std_par1<<", std_par2="<<a.std_par2<<endl;
    s<<"Lambert(CC): center_lon="<<a.lon_center<<", center_lat="<<a.lat_center;
    return s<<", false easting="<<a.false_easting<<", false northing="<<a.false_northing<<endl;
}

uint32 GProjLambert::getPSize() const
{
    return GMapProj::getPSize() + 
	sizeof(double)*2;  //std_par1 & std_par2
}

Result GProjLambert::input     (const char*  )
{
    return GOk;
}
Result GProjLambert::output    (char*& ) 
{
    return GOk;
}
Result GProjLambert::pack  (DataPipe& p)
{
    GMapProj::pack(p);
    p.append(CONPTR(&std_par1), sizeof(double));
    p.append(CONPTR(&std_par2), sizeof(double));
    return GOk;
}
Result GProjLambert::unpack    (DataPipe& p ) {
    GMapProj::unpack(p);
    GAssert(!p.eop());
    p.fetch((char*)&std_par1, sizeof(double));
    p.fetch((char*)&std_par2, sizeof(double));
    return GOk;
}

GProjLambert::GProjLambert(double r_maj, double r_min, 
			   double lat1, double lat2, 
			   double c_lon, double c_lat,
			   double x_false, double y_false)
{
    init(r_maj, r_min, lat1, lat2, c_lon, c_lat, x_false, y_false);
}

Result GProjLambert::init(double r_maj, double r_min, 
			  double lat1, double lat2, 
			  double c_lon, double c_lat,
			  double x_false, double y_false)
{
    double sin_po, cos_po;
    double ms1;                     // small m 1
    double ms2;                     // small m 2
    double ts0;                     // small t 0
    double ts1;                     // small t 1
    double ts2;                     // small t 2

    ellip.major = r_maj;
    ellip.minor = r_min;
    false_northing = y_false;
    false_easting = x_false;

    if (fabs(lat1+lat2) < EPSLN)  
	msgFail("LambertInit:: Equal Latitiudes for St. Parallels on opposite sides of equator");

    double temp = ellip.minor / ellip.major;
    es = 1.0 - temp*temp;
    e = sqrt(es);

    lon_center = c_lon;
    lat_center = c_lat;
    sin_po = sin(lat1), cos_po = cos(lat1);
    double con = sin_po;
    ms1 = ms(e,sin_po,cos_po);
    ts1 = ts(e,lat1,sin_po);
    sin_po = sin(lat2), cos_po = cos(lat2);
    ms2 = ms(e, sin_po, cos_po);
    ts2 = ts(e, lat2, sin_po);
    sin_po = sin(lat_center);
    ts0 = ts(e,lat_center,sin_po);
    if (fabs(lat1 - lat2) > EPSLN)
	ns = log (ms1/ms2)/ log (ts1/ts2);
    else
	ns = con;
    f0 = ms1 / (ns * pow(ts1,ns));
    rh = ellip.major * f0 * pow(ts0,ns);

    return GOk;
}


Result GProjLambert::forward(double lon, double lat, double& x, double& y)
{
    double con;       // temporary angle variable
    double rh1;       // height above ellipsoid
    double sinphi;    // sin value
    double theta;     // angle
    double _ts;        // small value t

    con  = fabs( fabs(lat) - HALF_PI);
    if (con > EPSLN) {
	sinphi = sin(lat);
	_ts = ts(e,lat,sinphi);
	rh1 = ellip.major * f0 * pow(_ts,ns);
    }
    else  {
	con = lat * ns;
	if (con <= 0) 
	    msgFail("Lambert::Forward: Point can not be projected");
	rh1 = 0;
    }
    theta = ns * adjustLon(lon - lon_center);
    x = rh1 * sin(theta) + false_easting;
    y = rh - rh1 * cos(theta) + false_northing;
    
    return GOk;
}

Result GProjLambert::inverse(double x, double y, double& lon, double& lat)
{
    double rh1;			// height above ellipsoid
    double con;			// sign variable
    double _ts;			// small t
    double theta;		// angle

    Result res = GOk;
    
    x -= false_easting;
    y = rh - y + false_northing;
    if (ns > 0) {
	rh1 = sqrt (x * x + y * y);
	con = 1.0;
    }
    else     {
	rh1 = -sqrt (x * x + y * y);
	con = -1.0;
    }
    theta = 0.0;
    if (rh1 != 0)
	theta = atan2((con * x),(con * y));
    if ((rh1 != 0) || (ns > 0.0)) {
	con = 1.0/ns;
	_ts = pow((rh1/(ellip.major * f0)),con);
	lat = phi2z(e,_ts, res);
	if (res != GOk)
	    return(res);
    }
    else
	lat = -HALF_PI;

    lon = adjustLon(theta/ns + lon_center);
    return GOk;
}



//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// Mercator projection

GProjMercator::GProjMercator()
{
    proj_id = Mercator;
}


GProjMercator::GProjMercator(double r_maj, double r_min, double c_lon, double c_lat,
			     double false_east, double false_north)
{
    init(r_maj, r_min, c_lon, c_lat, false_east, false_north);
}

Result  GProjMercator::init(double r_maj, double r_min, 
			    double center_lon, double center_lat,
			    double false_east, double false_north)
{
    proj_id = Mercator;
    ellip.major = r_maj;
    ellip.minor = r_min;
    lon_center = center_lon;
    lat_origin = center_lat;
    false_northing = false_north;
    false_easting = false_east;

    double temp = r_min / r_maj;
    es = 1.0 - (temp*temp);
    e = sqrt(es);
    m1 = cos(center_lat)/(sqrt(1.0 - es * sin(center_lat) * sin(center_lat)));
    return GOk;
}

Result  GProjMercator::forward(double lon, double lat, double& x, double& y)
{
    double _ts;		// small t value
    double sinphi;	// sin value

    if (fabs(fabs(lat) - HALF_PI)  <= EPSLN) {
	msgFail("Mercator::Forward: Transformation cannot be computed at the poles");
    } else  {
	sinphi = sin(lat);
	_ts = ts(e,lat,sinphi);
	x = false_easting + ellip.major * m1 * adjustLon(lon - lon_center);
	y = false_northing - ellip.major * m1 * log(_ts);
    }
    return(GOk);
}
    
Result  GProjMercator::inverse(double x, double y, double& lon, double& lat)
{
    double _ts;		// small t value
    Result res;		// error flag
    
    x -= false_easting;
    y -= false_northing;
    _ts = exp(-y/(ellip.major * m1));
    lat = phi2z(e,_ts,res);
    if (res != GOk)
	return(res);
    lon = adjustLon(lon_center + x/(ellip.major * m1));

    return(GOk);
}
    


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// Albers conical equal-area projection

GProjAlbers::GProjAlbers()
{
    proj_id = Albers;
}

GProjAlbers::GProjAlbers(double r_maj, double r_min, 
			 double lat1,  double lat2,
			 double lon0,  double lat0,
			 double fs_east, double fs_north)
{
    proj_id = Albers;
    init(r_maj, r_min, lat1, lat2, lon0, lat0, fs_east, fs_north);
}

Result	
GProjAlbers::init(double r_maj, double r_min, 
		  double lat1,  double lat2,
		  double lon0,  double lat0,
		  double fs_east, double fs_north)
{
    double sin_po, cos_po;
    double ms1;			// small m 1
    double ms2;			// small m 2
    double qs0;			// small q 0
    double qs1;			// small q 1
    double qs2;			// small q 2
    double con;

    std_par1 = lat1, std_par2 = lat2;
    ellip.major = r_maj;
    ellip.minor = r_min;
    
    lon_center = lon0;
    lat_center = lat0;
    false_northing = fs_north;
    false_easting = fs_east;
    if (fabs(lat1 + lat2) < EPSLN)   
	msgFail("Albers: invalid pair of standard parallels (equal distance from equator).");
    
    double temp = r_min/r_maj;
    es = 1.0 - temp*temp;
    e3 = sqrt(es);
    
    sin_po = sin(lat1), cos_po = cos(lat1);
    con = sin_po;
    ms1 = ms(e3, sin_po, cos_po);
    qs1 = qs(e3, sin_po, cos_po);
    
    sin_po = sin(lat2), cos_po = cos(lat2);
    ms2 = ms(e3, sin_po, cos_po);
    qs2 = qs(e3, sin_po, cos_po);
    
    sin_po = sin(lat0), cos_po = cos(lat0);
    qs0 = qs(e3, sin_po, cos_po);
    
    if( fabs(lat1-lat2) > EPSLN)
	ns0 = (ms1*ms1 - ms2*ms2)/(qs2-qs1);
    else
	ns0 = con;

    c = ms1*ms1 + ns0*qs1;
    rh = r_maj * sqrt(c-ns0*qs0)/ns0;
	return GOk;
}

Result
GProjAlbers::forward(double lon, double lat, double& x, double& y)
{
    double sin_phi,cos_phi;	// sine and cos values
    double _qs;			// small q
    double theta;		// angle
    double rh1;			// height above ellipsoid
    
    sin_phi = sin(lat), cos_phi = cos(lat);
    _qs = qs(e3, sin_phi, cos_phi);
    rh1 = ellip.major*sqrt(c-ns0*_qs)/ns0;
    theta = ns0*adjustLon(lon-lon_center);
    
    x = rh1*sin(theta) + false_easting;
    y = rh - rh1*cos(theta) + false_northing;
    
    return GOk;
}

Result
GProjAlbers::inverse(double x, double y, double& lon, double& lat)
{
    double rh1;			// height above ellipsoid
    double _qs;			// function q
    double con;			// temporary sign value
    double theta;		// angle
    Result flag;		// error flag
    double tx=x, ty=y;
    
    flag = GOk;
    tx -= false_easting;
    ty = rh - y + false_northing;
    
    if(ns0 >= 0) {
	rh1 = sqrt(tx*tx + ty*ty);
	con = 1.0;
    } else {
	rh1 = -sqrt(tx*tx + ty*ty);
	con = -1.0;
    }
    theta = 0.0;
    
    if (rh1 != 0.0)
	theta = atan2(con * tx, con * ty);
    con = rh1 * ns0 / ellip.major;
    _qs = (c - con * con) / ns0;
    if (e3 >= 1e-10) {
	con = 1 - .5 * (1.0 - es) * log((1.0 - e3) / (1.0 + e3))/e3;
	if (fabs(fabs(con) - fabs(_qs)) > .0000000001 )  {
	    lat = phi1z(e3, _qs, flag);
	    if (flag != GOk)
		return(flag);
	}
	else  {
	    if (_qs >= 0)
		lat = .5 * PI;
	    else
		lat = -.5 * PI;
	}
    }
    else   {
	lat = phi1z(e3, _qs, flag);
	if (flag != 0)
	    return(flag);
    }

    lon = adjustLon(theta/ns0 + lon_center);

    return(GOk);
}

ostream&    operator<<(ostream& s, const GProjAlbers& a)
{
    s<<"Albers: major="<<a.ellip.major<<", minor=";
    s<<a.ellip.minor<<", std_par1="<<GDEGREE(a.std_par1)<<", std_par2="<<GDEGREE(a.std_par2)<<endl;
    s<<"Albers: center_lon="<<GDEGREE(a.lon_center)<<", orgin_lat="<<GDEGREE(a.lat_center);
    return s<<", false easting="<<a.false_easting<<", false northing="<<a.false_northing<<endl;
}    
