/****************************************************************************
 *   tseries.cc : 
 *   		Implementation of TSeries class.
 *
 *   Change Log :
 *		 Jul-28-1997 Initial version,  Liujian Qian
 *
 *   $Id: gtseries.cpp,v 1.5 1998/12/10 00:11:46 qian Exp $
 ***************************************************************************/
 
#include "gtseries.h"


int 		GEpisodePool::n_total = 0;
int 		GEpisodePool::n_free = 0;
GEpisode* 	GEpisodePool::pool = 0;

GEpisode*
GEpisodePool::allocEpisode()
{
    GEpisode* ep;
    if(!pool) extendPool();
    ep = pool;
    if(!ep) return 0;
    pool = ep->next;
    ep->next = ep->prev = 0;
    n_free--;
    return ep;
}

// free an episode record and put it back to the pool
// episodes in the pool are only linked in a single direction.
void
GEpisodePool::freeEpisode(GEpisode* ep)
{	
    if(ep) {
	ep->next = pool;
	ep->prev = 0;
	memset(ep->evt_desc, 0, 8);
	pool = ep;
	n_free++;
    }
}


// extend the capacity of the pool w/ 'incr' of new episodes
Result
GEpisodePool::extendPool(int incr)
{
    GEpisode* ep;
    
    for(int i=0; i<incr; i++) {
	ep = ::new GEpisode;
	if(!ep) msgFail("can't extend episode pool!");
	if(!pool) pool = ep;
	else {
	    ep->next = pool;
	    pool = ep;
	}
	n_total ++, n_free ++;
    }
    return GOk;
}

// free all episodes; vacate the pool
void
GEpisodePool::freeAll()
{
    GEpisode *ep;
    
    ep = pool;
    while(ep) {
	pool = ep->next;
	ep->next=ep->prev=0;
	::delete ep;
	ep = pool;
    }
    pool = 0;
    n_total = n_free = 0;
}

ostream& operator<<(ostream& o, const GEpisode& e)
{
    return o<<(GTimestamp&)e;
}

GEpisode::GEpisode(const GEpisode& a) : GTimestamp(a)
{
    prev = next = 0;
    memcpy(evt_desc, a.evt_desc, 8);
}

void
GEpisode::setDate (int year, int mon, int day, int hour, int min, int sec)
{
    GTimestamp ts(year, mon, day, hour, min, sec);
    GTimestamp::operator=(ts);
}

GTimeSeries& 
GTimeSeries::operator= (const GTimeSeries& a)
{
    GEpisode *one, *ep;
    
    one = a.firstEpisode();
    while(one) {
	ep = new GEpisode((const GEpisode&)(*one));
	addEpisode(ep);
	one = one->nextEpisode();
	if(one == a.firstEpisode()) break;
    }
    return *this;
}


GTimeSeries::~GTimeSeries()
{
    if(deleteAllEpisodes() != GOk) 
	msgWarn("Error deallocating all episodes!");
}

	
    
// add an episode into the series (in sorted order)
// assumming smallest one is the first one.
Result
GTimeSeries::addEpisode(GEpisode* ep)
{		
    GEpisode *cur;
    
    if (!first) { // *ep is the first episode
	first = ep;
	first->next = first->prev = first;
	last_active = ep;
	n_episodes ++;
	return GOk;
    }
    if(*ep < *first) { // *ep smller than anyone
	first->prev->next = ep;
	ep->next = first;
	ep->prev = first->prev;
	first->prev = ep;
	first = ep;
	last_active = ep;
	n_episodes ++;
	return GOk;
    }
    
    if(n_episodes > 5 && last_active) {
	//try to check the 'last_active' episode; the idea
	//is that we may want to add 'ep' just before or after
	//'last_active', particularly if we are adding a lot
	//of adjacent episodes.
	if(*ep<=*last_active && *ep >= *(last_active->prev)) {
	    //add ep before last_active
	    ep->prev = last_active->prev;
	    ep->next = last_active;
	    last_active->prev->next = ep;
	    last_active->prev = ep;
	    n_episodes ++;
	    last_active = ep;
	    return GOk;
	}
	else if( (*ep>=*last_active && *ep<=*(last_active->next)) ||
		 (*ep>=*last_active && last_active->next == first) ) {
	    //add ep after last_active
	    ep->next = last_active->next;
	    last_active->next->prev = ep;
	    ep->prev = last_active;
	    last_active->next = ep;
	    n_episodes ++;
	    last_active = ep;
	    return GOk;
	}
    }

    // do a brute force search if gets here
    cur  = first;
    while(*cur<*ep) {
	cur = cur->next;
	if(cur==first) break;
    }
    // insert 'ep' before 'cur'
    cur->prev->next = ep;
    ep->next = cur;
    ep->prev = cur->prev;
    cur->prev = ep;
    n_episodes ++;

    last_active = ep;
    return GOk;
}


// delete an episode from the series (and put back to the pool)
Result
GTimeSeries::delEpisode(GEpisode* ep)
{
    ep->next->prev = ep->prev;
    ep->prev->next = ep->next;
    ep->prev= ep->next = 0;
    delete ep;
    return GOk;
}


Result
GTimeSeries::reset(TimeUnit u)
{
    if(u!=_INVALID) t_unit = u;
    return deleteAllEpisodes();   
}

Result
GTimeSeries::deleteAllEpisodes(void)
{
    if(first) {
	GEpisode *cur = first->next, *next;
	while(cur != first) {
	    if(!cur) break;
	    next = cur->next;
	    delete cur;
	    cur = next;
	}
	delete first;
    }
    first = 0;
    n_episodes = 0;
    last_active = 0;
    return GOk;
}


// add a series of episodes begining with 'from', increasing 
// with the t_unit, and until readches timestamp 'to'.
Result
GTimeSeries::addRange(const GDate& from, const GDate& to, int delta)
{
    GTimestamp begin(from), end(to);
    
    return addRange(begin, end, delta);
}

Result
GTimeSeries::addRange(const GTimestamp& begin, const GTimestamp& end, int delta)
{
    GEpisode * new_ep =0;
    GTimestamp ts = begin;
    long sec_delta;
    
    if(t_unit==_INVALID || t_unit>= _MONTH) 
	msgFail("addRange not supported for this time unit!");
    sec_delta = GTimestamp::_seconds[t_unit];

    while(ts<=end) {
	new_ep = new GEpisode;
	if(!new_ep) return GFail;
	new_ep->setTimestamp(ts);
	addEpisode(new_ep);
	ts += sec_delta*delta;
	//cout<<"addRange: next timestamp: "<<ts<<endl;
    }
    
    return GOk;
}    


// generalize the series to a larger unit (eg, hours -> days)
Result
GTimeSeries::generalize(TimeUnit )
{
    return GOk;
}


ostream&    
operator<<(ostream& os, const GTimeSeries& t)
{    os<<"total # of episodes = "<<t.numEpisodes()<<endl;
    os<<"unit of episodes = "<<t.timeUnit()<<endl;
    GEpisode *ep = t.firstEpisode();
    os<<(GTimestamp&)*ep;
    while( (ep = ep->nextEpisode()) ) {
	if(ep == t.firstEpisode() ) break;
	os<<"\n"<<(GTimestamp&)*ep;
    }
    return os<<endl;
}

    
