/****************************************************************************
 *   rt.cpp : 
 *   		implementation file for RTree class.
 *
 *   Change Log :
 *		  09-17-1997 Initial version,  Liujian Qian
 *
 *   $Id: rt.cpp,v 1.5 1998/12/10 00:11:47 qian Exp $
 ***************************************************************************/
#include "iostream.h"
#include "rt.h"

// reset the internal rtree cursor
void
RCursor_i::init()
{
    stack.reset();
    path.reset();
    
    no_more 	= 1;
    pred.op 	= RPred::BAD;
    first   	= 1;
    last_level 	= -1;
    n_records   = 0;
}


//////////////////////////////////////////////////////////////////////

int	RTree::cur;

///
RTree::RTree() : max_cursor(32)
{
    is_open = 0;
    store   = 0;
    cursor_i = new RCursor_i[max_cursor];
    GAssert(cursor_i);
}

RTree::~RTree()
{
    if(is_open) close();
    if(cursor_i) delete[] cursor_i;
}


ostream&	
operator<<(ostream& os, const RTree& r)
{
    return os<<*(r.store->rtHeader());
}

Result
RTree::create(const char* name)
{
    PageID  pid;

    if (is_open) 
	msgFail("rtree still open");

    newStore();

    if( store->create(name) != GOk)
	msgFail("can't create store");
    
    pid = store->allocPage();

    RNode *node = newNode();

    // initialize the first node as root node
    node->setPointer( RPtr(RT_ROOT_PAGE) );
    node->setLevel(0);
    
    writeNode(node);

    delete node;

    is_open = 1;
    return GOk;
}


Result
RTree::open (const char* name)
{
    if (is_open) 
	msgFail("rtree already open");

    store = newStore();
    store->open(name);
    if (!store->isOpen())
	msgFail("can't open store file");

    is_open = 1;
    return GOk;
}
	
Result
RTree::close()
{
    if (is_open) {
	store->close();
	is_open = 0;
	delete store;
	return GOk;
    }    
    else msgFail("RTree is not open");
}

RTHdr*	
RTree::stats()
{
    return store->rtHeader();
    //!!! add display of the average node occupancy here !!! 
}

    

RTStore*
RTree::newStore    () 
{
    store = new RTStore;
    return store;
}

RNode*
RTree::newNode     () 
{
    return new RNode;
}

RNode*	
RTree::readNode    (const RPtr& p)
{
    char* page=0;
    DataPipe pipe;

    if(!p.id()) {
	msgWarn("invalid page id given");
	return 0;
    }
    RNode *node = newNode();
    store->read(p.id(), page);

    pipe.attach(page, RTPAGE_SZ);
    pipe.seek(0);
    node->unpack(pipe);
    
    node->setPointer(p);
    return node;
}

void
RTree::writeNode   (RNode* nd)
{
    store->cleanBuf();
    DataPipe pipe;
    pipe.attach( store->_buf, RTPAGE_SZ);
    pipe.recycle();

    //pack all the stuff in the node into the store buffer page
    nd->pack(pipe);
    //then write into the permanent store
    store->write(nd->pointer().id(), store->_buf);
}


////////////////////////////////////////////////////////////////////////

// this decends down starting from 'pg', till it reaches a
// leaf, where the insertion of e causes minmal penalty 
// (all along the way).
//
RNode*		
RTree::pickLeaf (const RPtr& p, const REntry& e, int level)
{
    RNode* 	node=0;
    RPtr	ptr;
    RPtrStk	path;
    
    ptr = p;
    path.push(ptr);
    
    for(;;) {
	node = readNode(ptr);
	if( level == node->level() || node->isLeaf() ) {
	    node->setPath(path);	//keep track of the descending path
	    break;
	}
	// based on penalty calculation, decends along best path
	ptr = node->searchMinPenalty( e );
	path.push(ptr);
	delete node;
    }    
    return node;
}


Result
RTree::insert (const REntry& e)
{
    if(store->hdr.n_levels==0) store->hdr.n_levels = 1;
    return _insert(e, 0);
}

// this does the real work
Result
RTree::_insert(const REntry& e,	//IN, the incomming entry
	       int	level)		//IN, which level should e go
{
    RNode* 	leaf, *old;
    int		overflow = 0;
    
    // setp1: find out the best place e should go
    leaf = pickLeaf(RPtr(RT_ROOT_PAGE), e, level);
    leaf->insert(e);
    old = leaf;

    // step2: split the leaf node if its overfull;
    //        this also propagate chagnes upwards
    if(leaf->isOverFull()) {
	split(&leaf, e);
    }
    else 
	writeNode(leaf);
	
    // step3: 
    if(!overflow)
	//new rect inserted; the key in parent's entry that
	//contains this node should be updated/adjusted, which
	//can also propagate upwards to root.
        adjustKeys(leaf, 0);
    
    delete leaf;
    
    return GOk;
}
    
// the node nd, after e has been inserted into it, is now overfull.
// this method will split nd into two nodes and updates entries in
// the parent node... after split the node where new entry 'e' end
// up with is returned. *This is a recursive method*
// 
Result
RTree::split 	   (RNode** nd, const REntry& e)
{
    bool	went_left =0, new_root = 0;
    RPtrStk	path;

    //cout<<"spliting node="<<(*nd)->pointer()<<endl;
    
    //step 1: check if spliting root
    if( (*nd)->isRoot()) {
	new_root  = 1;
	(*nd)->setPointer( RPtr(store->allocPage()) );
	store->hdr.n_levels ++;
    }
    
    path = (*nd)->path();
    
    //step 2: let the node split; node2 is the right-hand node
    RNode* node2 = (*nd)->split();
    node2->setPointer ( RPtr(store->allocPage()) );

    int pos;
    //see where the new entry went
    if( (*nd)->searchPtr(e.pointer(), pos) )
	went_left = 1;

    //link to old node's sibling
    node2->setSibling ( (*nd)->sibling() );
    (*nd)->setSibling ( node2->pointer() );

    writeNode(*nd);
    writeNode(node2);
    
    // generate two new entries for parent node
    REntry *ep1 = (*nd)->unionIt();
    REntry *ep2 = node2->unionIt();
    ep1->setPointer ( (*nd)->pointer() );
    ep2->setPointer ( node2->pointer() );
    if (new_root) {     // lift the tree if spliting root
	RNode *root = newNode();
	root->setLevel ( (*nd)->level() + 1 );
	root->insertBefore(*ep1, 0);
	root->insertBefore(*ep2, 1);
	root->setPointer ( RPtr(RT_ROOT_PAGE) );
	writeNode(root);
	// update the path in nd and node2
	path.reset();
	path.push(RPtr(RT_ROOT_PAGE));
	path.push(ep1->pointer());
	(*nd)->setPath(path);
	path.pop();
	path.push(ep2->pointer());
	node2->setPath(path);
	delete root;
	
    } else {	// insert entry for N' in parent

	//find the parent node first
	RPtr parent_ptr = path.second();

	/// XXX this readNode is causing side-effect on ep2 XXX
	RNode  *parent   = readNode(parent_ptr);
	
	path.pop();
	parent -> setPath(path);

	// find the original entry in parent
	RPtr   ptr;
	int    pos;
	ptr   =  (*nd)->pointer() ;

	REntry* ep = parent->searchPtr (ptr, pos);
	GAssert (ep);
	
	// update/insert new entries in parent node	
	parent->remove(pos);
	parent->insertBefore(*ep1, pos);
	parent->insertBefore(*ep2, pos+1);
	// see if parent needs split too
	if( !parent->isOverFull() )  //we are lucky
	    writeNode( parent );
	else {	//recursively call split against parent
	    split(&parent, went_left?*ep1:*ep2);

	    // adjust the paths to reflect the parental change  
	    RPtr p = (*nd)->pointer();
	    path   = parent->path();
	    path.push(p);
	    (*nd)->setPath(path);
	    
	    p = node2->pointer();
	    path.pop();
	    path.push(p);
	    node2->setPath(path);
	}
	
	delete parent;
    } //else
    
    if (!went_left) {
	delete *nd;
	*nd = node2;
    } else 
	delete node2;
	
    delete ep1;
    delete ep2;

    return GOk;
}


//
// nd's path from root down to itself 
// should have been set up. update the key
// in the parent node
Result		
RTree::adjustKeys (RNode* nd, RNode **parent)
{
    RPtrStk	path;

    path	= nd->path();

    if(nd->isRoot() || path.size()<=1 ) return GOk;

    RPtr	me= path.pop();
    if(path.isEmpty()) return GOk;
    
    RPtr	pparent = path.top();    
    RNode	*p;
    if(parent)
	p = *parent;
    else {
	p = readNode(pparent);
	parent = &p;
    }
	
    GAssert(p);

    // find out the entry that points to 'nd'.
    int 	pos;
    REntry	*e = p->searchPtr(me, pos);
    GAssert(e);
    
    REntry	*eu = nd->unionIt();
    eu->setPointer (e->pointer());
    
    if( !eu->equal(*e) ) {
	p->remove(pos);
	p->insertBefore(*eu, pos);

	// !!! potential glitch !!!
	if(p->isOverFull()) msgWarn("reinsert causing overfull!");

	writeNode(p);
	p->setPath(path);	//setup its own path
	adjustKeys(p, 0);
    }
    
    if (parent == &p) delete p;
    
    delete eu;
    //delete e;	//this has/will been deleted by the owning node

    return GOk;
}

//
int
RTree::condenseTree(RNode* leaf)
{
    if(!leaf) return -1;
    else return 0;
}


//
Result
RTree::shortenTree ()
{
    return GOk;
}

//
// set the internal cursor structure and return an integer 
// to user as the client-end cursor.
int	
RTree::search (const RPred& p)
{
    GAssert(is_open);

    // !!! should add mutex lock here if multi-threaded !!!

    for( int i=0; i<max_cursor; i++)     //find a free cursor_i slot
	if(! cursor_i[i].in_use) {
	    cursor_i[i].init();
	    cursor_i[i].pred = p;
	    cursor_i[i].in_use = 1;
	    return i;
	}

    msgWarn("no more cursors available!");
    return -1;
}


REntry*	
RTree::next   (int c)
{
    RPtr	ptr;
    REntry 	*entry;
    
    if(c <0 || c>=max_cursor) {
	msgWarn("Invalid cursor used for getting next record");
	return (REntry*)0;
    }

    while(cursor_i[c].first || !cursor_i[c].stack.isEmpty() ) {
	if(cursor_i[c].first) {
	    ptr.setId(RT_ROOT_PAGE);
	    cursor_i[c].first = 0;
	} else {
	    entry = new REntry;
	    *entry= cursor_i[c].stack.pop();
	    if(entry->pointer().isLeaf()) {
		cursor_i[c].n_records++;
		return entry;
	    }
	    ptr = entry->pointer();
	    delete entry;
	}
	
	RNode *node = readNode(ptr);
	GAssert(node);

	//push all the satisfying entries from 'node' into stack
	GList<REntry*> *list = node->search(cursor_i[c].pred);
	int  idx;
	REntry **ep = list->first(&idx);
	while(ep) {
	    entry = *ep;
	    if(node->isLeaf()) entry->setLeaf(true);
	    cursor_i[c].stack.push(*entry);
	    ep = list->next(&idx);
	}

	delete list;
	delete node;
	
    } //while

    //no more good entries; implicitly close the cursor and free it
    cursor_i[c].init();
    cursor_i[c].in_use = 0;
    return (REntry*) 0;
}


Result
RTree::bulkLoad (const char* file)
{
    if(!file)
	msgFail("invalid file name");
    
    return GOk;
}

    
Result
RTree::remove (const RPred& )
{
    return GOk;
}

void
RTree::printAll()
{
    _print(RPtr(RT_ROOT_PAGE));
}


void
RTree::_print(const RPtr& p )
{
    //REntryStk	stack;
    REntry	e;
    RNode	*node;
    RPtr	ptr;

    node = readNode(p);
    if(node) {
	cout<<*node;
	if(!node->isLeaf()){
	    for(int i=0; i<node->numEntries(); i++) {
		ptr = node->entries[i]->pointer();
		_print(ptr);
	    }
	}
	delete node;
    }

    return;
}

