/****************************************************************************
 *   rtentry.cc : 
 *   		 implementation file for Rtree Entry class.
 *	
 *   Change Log :
 *		 09-17-1997 Initial version,  Liujian Qian
 *
 *   $Id: rtentry.cpp,v 1.3 1998/12/02 04:03:44 qian Exp $
 ***************************************************************************/

#include "rtentry.h"


RPtr& 	
RPtr::operator= (const RPtr& a) 
{
    is_remote = a.is_remote;
    is_leaf   = a.is_leaf;
    host_id   = a.host_id;
    the_id    = a.the_id;
    return *this;
}

bool
RPtr::operator==(const RPtr& a) const
{
    return is_remote==a.is_remote && host_id==a.host_id && 
	   the_id==a.the_id;
}

ostream&	operator<<(ostream& os, const RPtr& p)
{
    os<<"remote="<<(int)p.is_remote;
    return  os<<",hostID="<<p.host_id<<",rec/pageID="<<p.the_id;
}


//////////////////////////////////////////////////////////////////////

REntry::~REntry() 
{
    //init();
}

REntry& 
REntry::operator= (const REntry& a) 
{
    //cout<<"REntry::operator=, this="<<*this<<",other="<<a<<endl;
    _key   = a._key,    ptr = a.ptr;
    _level = a._level;
    next  = a.next;
    return *this;
}


void
REntry::init() 
{
    _level = 0, _key.set(-1,-1,-1,-1);
    ptr.init(); next = 0;
}        

void * 		
REntry::operator new(size_t)
{
    return REntryPool::allocEntry();
}

void
REntry::operator delete(void* ep, size_t) 
{
    REntryPool::freeEntry((REntry*)ep); 
}

ostream&	
operator<<(ostream& os, const REntry& e)
{
    return os<<"key="<<e.key()<<", ptr="<<e.pointer().id();
}


int
REntry::equal 	   (const REntry& other) const
{
    //there is not much we can do here regarding 
    //general RTree ... 
    if(_key.equals(other._key) && ptr.operator==(other.ptr) ) 
	return 1; 
    else 
	return 0; 
}

int 
REntry::cmp	   (const REntry& other) const
{
    //change this if you want the entries be sorted differently
    //when inserting them.
    return _key.cmp(other._key);
}

int
REntry::hCmp	   (const REntry& other) const
{
    long my_h = hOrder(), other_h=other.hOrder();

    return (my_h>other_h)?1:(my_h==other_h)?0:-1;
}

bool
REntry::consistent (const RPred& pred)
{
    switch (pred.op) {
    case RPred::EQUALS:
	return  _key.cmp(pred.window) == 0;

    case RPred::OVERLAPS:
	return  _key.intersects(pred.window);

    case RPred::CONTAINS:
	return  _key.covers(pred.window);
	
    case RPred::WITHIN:
	return  pred.window.covers(_key);
	
    default:
	msgWarn("invalid operator in the predicate");
	return false;
    }
}


bool
REntry::operator== (const REntry& other)
{
    return ( equal(other) );
}

bool
REntry::overlaps   (const REntry& other) const
{
    return _key.intersects(other._key);
}

bool
REntry::contains   (const REntry& other) const
{
    return _key.covers(other._key);
}

bool
REntry::isInside   (const REntry& other) const 
{
    return other.contains(*this);
}


RPenalty
REntry::penalty	(const REntry& other) const 
{
    RKey tmp;
    
    tmp = _key;
    tmp.enlarge(other.key() );
    return  GMAX( (tmp.area() - other.key().area() ), 0 );
}

Result
REntry::pack	(DataPipe& p) 
{
    if( _key.pack(p) != GOk) 
	msgFail("can't pack RKey");
    
    if( ptr.pack(p) != GOk)
	msgFail("can't pack RPtr");
    
    return GOk;
}
    
Result
REntry::unpack  (DataPipe& p)
{
    if( _key.unpack(p) != GOk) 
	msgFail("can't unpack RKey");
    
    if(ptr.unpack(p) != GOk)
	msgFail("can't unpack RPtr");
    return GOk;
}

    
int
REntry::getPSize()
{
    return _key.getPSize()+ptr.getPSize();
}


int 	REntryPool::n_total=0;
int	REntryPool::n_free=0;
REntry* REntryPool::pool=0;

//////////////////////////////////////////////////////////////////////
REntry* 	
REntryPool::allocEntry()
{
    REntry* ep=0;

    if(n_free<=0) extendPool();
    ep = REntryPool::pool;
    if(!ep) {
	stats();
	msgWarn("can't get new entry from pool; pool corrupted?");
	return 0;
    }
    
    pool = ep->next;
    ep->next  = 0;
    n_free--;
    return ep;
}

void	
REntryPool::freeEntry(REntry* ep)
{
    if(ep) {
	//sanity checking: XXX this is taking 26% of total time XXX
	/*
	while(tmp) {
	    if(tmp == ep) {
		msgWarn("entry being deleted for second time!");
		return;
	    }
	    tmp = tmp->next;
	}
	*/
	
	ep->next = pool;
	pool = ep;
	n_free++;
	//cout<<"after free: pool="<<(int)pool<<endl;
    }
}

Result
REntryPool::extendPool(int incr) 
{
    REntry* ep;
    
    for(int i=0; i<incr; i++) {
	ep = ::new REntry;
	if(!ep) msgFail("can't extend entry pool!");
	if(!pool) pool = ep;
	else {
	    ep->next = pool;
	    pool = ep;
	}
	n_total ++, n_free ++;
    }    
    return GOk;
}

void
REntryPool::freeAll()
{
    REntry *ep;
    
    ep = pool;
    while(ep) {
	pool = ep->next;
	ep->next=0;
	::delete ep;
	ep = pool;
    }
    pool = 0;
    n_total = n_free = 0;
    
}


ostream&	operator<<(ostream& os, const RPtrStk& t)
{
    os<<"(stk:";
    for(int i=0; i<t._top; i++) {
	os<<t._stk[i].id()<<", ";
    }
    os<<")";
    return os;
}

