/****************************************************************************
 *   RTNode.cc : 
 *   		implementation file for Rtree Node class.
 *
 *   Change Log :
 *		  09-17-1997 Initial version,  Liujian Qian
 *
 *   $Id: rtnode.cpp,v 1.4 1998/12/02 04:03:45 qian Exp $
 ***************************************************************************/

#ifdef __GNUG__
#pragma implementation "rtnode.h"
#endif

#include "rtnode.h"

//these are required for proper instantiate of templates by GNU C++.
#ifdef __GNUC__
template class GList<REntry*>;
template class Dlink<REntry*>;
#endif

///
RNode::RNode ()
{
    memset(entries, 0, MAX_ETS*sizeof(REntry*));
    hdr.n_entries = hdr.level = 0;    
}

///
RNode::RNode (const RNode& node)
{ 
    REntry* ep;

    hdr.n_entries = node.hdr.n_entries;
    hdr.level	  = node.hdr.level;
    hdr.sibling	  = node.hdr.sibling;
    
    memset(entries, 0, MAX_ETS*sizeof(REntry*));
    //do the real copying
    for(int i=0; i<numEntries(); i++) {
	ep = new REntry;
	*ep = *(node.entries[i]);
	entries[i] = ep;
    }	
}

///
RNode::~RNode()
{
    for(int i=0; i<numEntries(); i++)
	if(entries[i]) delete entries[i];
    
    //RNodePool::stats();
}

///
void * 	RNode::operator new(size_t)
{
    return RNodePool::allocNode();
}

///
void  	RNode::operator delete(void* nd, size_t)
{
    RNodePool::freeNode( (RNode*) nd );
}


///
bool
RNode::isOverFull () const
{
    bool yes = getPSize() > RTPAGE_SZ;
    if(yes) {
	//cout<<"page "<<pointer().id()<<" is overfull"<<endl;
	//cout<<",level="<<level()<<", parent="<<parent()<<endl;
    }
    return yes;
}

///
bool
RNode::isUnderFull () const
{
    return getPSize() < RTPAGE_SZ/2;
}

///
int
RNode::getPSize  () const
{
    if(numEntries()<=0) {
	msgWarn("empty rtree node?");
	return sizeof(NodeHdr);
    }
    
    int e_sz = entries[0]->getPSize();
    
    return 	sizeof(NodeHdr) + numEntries()*e_sz;
}

Result
RNode::pack (DataPipe& p)
{
    p.recycle();
    p.append((char*)&hdr, sizeof(NodeHdr));
    for(int i=0; i<numEntries(); i++)
	entries[i]->pack(p);

    return GOk;
}

Result
RNode::unpack(DataPipe& p)
{
    //p.seek(0);
    if( p.fetch((char*)&hdr, sizeof(NodeHdr)) !=GOk )
	msgFail("can't fetch node header");
    //cout<<"node after unpacking header:"<<*this;
    for(int i=0; i<numEntries(); i++) {
	entries[i] = new REntry;
	entries[i]->unpack(p);
	entries[i]->setLevel(level());
    }

    return GOk;
}

ostream&
operator<<(ostream& os, const RNode& n)
{
    for(int j=0; j<4-n.level(); j++) os<<"\t";	
    os<<"*level="<<n.level()<<", pgid="<<n.pointer().id();
    os<<", #entries="<<n.numEntries()<<"\n";
    //os<<", sibling="<<n.sibling()<<"\n (";
    for(int i=0; i<n.numEntries(); i++){
	for(int j=0; j<4-n.level(); j++) os<<"\t";	
	os<<"    "<<*n.entries[i];
	os<<" \n";
    }
    return os; 	//<<")\n";
}

REntry*	
RNode::getEntry(int i) const
{
    if(i>=0 && i<hdr.n_entries) 
	return (REntry*) entries[i];
    else return (REntry*)0;
}

	
GList<REntry*>*	
RNode::search(const RPred& query) const
{
    GList<REntry*> *list = new GList<REntry*>;
    GAssert(list);
    
    for(int i=0; i<numEntries(); i++) {
	if(entries[i]->consistent(query))
	    list->append(entries[i]);
    }
    
    return list;
}

REntry*
RNode::searchPtr (const RPtr& ptr, int& pos) const
{
    //cout<<"search for pointer:"<<ptr<<endl;
    
    for(int i=0; i<numEntries(); i++) {
	if(!entries[i]) {
	    cout<<"entries "<<i<<" is invalid; node:"<<*this;
	    break;
	}
	// for now we only compare the page/record id;
	// things such as remote host ID are NOT compared
	if( entries[i]->pointer().id()==ptr.id() ) {
	    pos = i;
	    return entries[i];
	}
    }
    return 0;
}

//
// loop through all the entries in this node, and
// look for the entry that will cause minimum penalty 
// (enlargement in MBR) if 'e' is to be inserted down 
// its subtree.
//
RPtr
RNode::searchMinPenalty (const REntry& e) const
{
    RPtr 	pid;
    REntry 	*ep;
    RPenalty	p=-1.0, min_p=-1.0;
    
    for(int i=0; i<numEntries(); i++) {
	ep = entries[i];
	GAssert(ep);
	p = ep->penalty(e);
	if(min_p <0 || min_p > p) {
	    min_p = p;
	    pid = ep->pointer();
	}
    }
    
    return pid;
}


PageID
RNode::searchNeighbors (PageID ) const
{
    return -1;   
}

    
Result
RNode::insert(const REntry& e)
{
    int i;
    //inovke the entry's compare function and determin	
    //where should the new entry go
    for(i=0; i<numEntries(); i++)
	if (entries[i]->cmp(e) >0)
	    break;
    
    return insertBefore(e, i);
}

//
// this will copy the  the new entry and 
// store it (not the pointer!) in the node
//
Result
RNode::insertBefore(const REntry& e, int idx)
{	
    GAssert(idx<=numEntries());
    REntry *ep = new REntry;
    
    *ep = e;
    ep->setLevel(level());
    //cout<<"node::insertBeofre="<<idx<<",entry:"<<*ep;
    
    //shuffle things around
    for (int i=numEntries(); i>idx; i--)
	entries[i] = entries[i-1];
    entries[idx] = ep;
    hdr.n_entries ++;
    
    return GOk;
}


Result
RNode::remove(int idx)
{
    int i;
    
    if(!entries[idx])
	msgFail("deleting non-existing entry");
    delete entries[idx];
    for(i=idx; i<numEntries()-1; i++)
	entries[i] = entries[i+1];
    hdr.n_entries--;

    return GOk;
}

Result
RNode::bulkRemove (int vec[], int num)
{
    int i,j;
    REntry* new_entries[MAX_ETS];
    
    memset(new_entries, 0, sizeof(REntry*)*MAX_ETS);
    for(i=0; i<num; i++) {
	delete entries[vec[i]];
	entries[vec[i]] = 0;
    }
    for(i=0, j=0; i<numEntries(); i++) 
	if(entries[i])
	    new_entries[j++] = 	entries[i];
    GAssert( numEntries()-j == num);
    hdr.n_entries = j;
    
    //now copy new_entries back into old entries so it leaves no gap
    for(i=0; i<numEntries(); i++)
	entries[i] = new_entries[i];
    
    return GOk;
}

//
// this loops through all entries and generate a
// new key (MBR) that encloses all the rectangles of 
// the entries.
//
REntry*	
RNode::unionIt () const
{
    REntry*	ep = new REntry;
    RKey	new_key;

    new_key = entries[0]->key();
    for(int i=1; i<numEntries(); i++)
	new_key.enlarge(entries[i]->key());

    ep->setKey(new_key);
    
    return ep;
}

//
// Guttman's poly-time split algorithm is used
// in spliting the entries in one node into two
// groups of entries that are to be put in two nodes.
// After split, the origin node becomes the left node;
// another node is the right node which will be returned.
//
RNode*	
RNode::split	()
{
    int 	i, j, right_dels=0, left_dels=0, 
	        seed1=-1, seed2=-1;
    int		right_vec[MAX_ETS], left_vec[MAX_ETS];
    double 	max= -1.0, dist;
    REntry 	*tmp_ep1, *tmp_ep2, *ep;
    RKey	cover1, cover2, key1, key2;

    memset(right_vec, 0, MAX_ETS*sizeof(int));
    memset(left_vec,  0, MAX_ETS*sizeof(int));
    
    // step1: pick two entries that are mostly apart
    // 	      cost: O(N^2) 
    for(i=0; i<numEntries(); i++) {
	tmp_ep1 = entries[i];
	for(j=i+1; j<numEntries(); j++) {
	    tmp_ep2 = entries[j];
	    dist = tmp_ep1->key().distance (tmp_ep2->key());
	    if(dist > max) {
		max   = dist;
		seed1 = i;
		seed2 = j;
	    }
	}
    }
    
    tmp_ep1 = entries[seed1], tmp_ep2 = entries[seed2];
    cover1  = tmp_ep1->key(), cover2  = tmp_ep2->key();
    
    // step2: duplicate this node for right node
    RNode*  right_node = new RNode (*this);
    
    // setp3: mark appropriate entries for deletion from the nodes
    for(i=0; i<numEntries(); i++) {
	if(i==seed1) //seed1 belong in left node
	    right_vec[right_dels++] = i;
	else if(i==seed2)  //seed2 belong in right node
	    left_vec[left_dels++] = i;
	else {
	//determin where should this entry go based on the penalty
	//wrt the seed in each node
	ep = entries[i];
	key1 = cover1, key2 = cover2;
	key1.enlarge(ep->key());
	key2.enlarge(ep->key());
	if( key2.area() > key1.area() ) 
	    left_vec[left_dels++] = i;
	else
	    right_vec[right_dels++] = i;
	}
    }
    
    // step4: do the actual deletions 
    bulkRemove (left_vec, left_dels);
    right_node->bulkRemove(right_vec, right_dels);
    
    return right_node;
}


int 	RNodePool::n_total;
int	RNodePool::n_free;
RNode*  RNodePool::pool=0;

//////////////////////////////////////////////////////////////////////
RNode* 	
RNodePool::allocNode()
{
    RNode* ep;
    if(n_free<=0) extendPool();
    ep = pool;
    if(!ep) return 0;
    pool = ep->next;
    ep->next  = 0;
    n_free--;
    return ep;
}

void	
RNodePool::freeNode(RNode* ep)
{
    if(ep) {
	ep->next = pool;
	pool = ep;
	n_free++;
    }
}

Result
RNodePool::extendPool(int incr) 
{
    RNode* ep;
    
    for(int i=0; i<incr; i++) {
	ep = ::new RNode;
	if(!ep) msgFail("can't extend node pool!");
	if(!pool) pool = ep;
	else {
	    ep->next = pool;
	    pool = ep;
	}
	n_total ++, n_free ++;
    }    
    cout<<"****************************** Node Pool Extended  ******************************\n";
    return GOk;
}

void
RNodePool::freeAll()
{
    RNode *ep;
    
    ep = pool;
    while(ep) {
	pool = ep->next;
	ep->next=0;
	::delete ep;
	ep = pool;
    }
    pool = 0;
    n_total = n_free = 0;
    
}
