/****************************************************************************
 *   rtstore.cc : 
 *   		 implementation file for RTree Store class.
 *
 *   Change Log :
 *		 09-17-1997 Initial version,  Liujian Qian
 *
 *   $Id: rtstore.cpp,v 1.4 1998/12/02 04:03:46 qian Exp $
 ***************************************************************************/
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#if defined(__unix__) || defined (__unix)
#include <unistd.h>
#else
#include <stdio.h>
#endif

#include "rtstore.h"


// page 0 in an rtree starts whith these  magic words, followed
// by the head of free page list, followed by the RTree header.
//
static char	rt_magic[] = "GFC RTree Index";

ostream&	
operator<<(ostream& os, const RTHdr& h)
{
    os<<"#levels="<<h.n_levels<<", #nodes="<<h.n_nodes;
    os<<", #entries="<<h.n_entries<<", mbox=";
    return os<<GRect(h.x_min, h.y_min, h.x_max, h.y_max)<<endl;
}


RTStore::RTStore () : is_open(0) 
{
    hdr.init(); 
    cleanBuf();
    root_node = 0;
} 	    	


Result
RTStore::create(const char* name)
{
    if( isOpen() ) {
	msgWarn("closing previously open index");
	close();
    }
    
    
    file_handle = ::open (name, O_RDWR | O_CREAT 
			| O_TRUNC, S_IREAD | S_IWRITE);
    if(file_handle<0)
	msgFail("can't create new store");
    
    setOpen(1);

    //initialize    the rtree header
    hdr.init();
    
    //initialize first (reserved) page, whose page id is always 0!
    //
    cleanBuf();
    memcpy(_buf,  rt_magic, sizeof(rt_magic));
    memcpy(_buf+sizeof(rt_magic)+sizeof(PageID), &hdr, sizeof(RTHdr));
    ::write(file_handle, _buf, RTPAGE_SZ);
    cleanBuf();

#   ifdef  USE_MMAP
    hdr.map_len = RTPAGE_SZ;
    hdr.map_region = mmap(NULL, hdr.map_len, PROT_READ|PROT_WRITE,
		      MAP_SHARED, file_handle, (off_t)0 );
    if(hdr.map_region == (caddr_t)-1 ) {
	msgWarn("can't mmap the rtree");
	::close(file_handle);
	return GFail;
    }
#   endif
    
    return GOk;
}

Result
RTStore::open(const char* name)
{
    if(isOpen()) 
	msgFail("must close opening index first");
    
    file_handle = ::open(name, O_RDWR);
    if (file_handle < 0)
	msgFail("can't open index file");

    char *page = _buf;

    ::read(file_handle, (char*)page, RTPAGE_SZ);
    if(memcmp(page, rt_magic, sizeof(rt_magic))){
	msgWarn("can't find the magic word: rtree file corrupted?");
	//::close(file_handle);
	//return GFail;
    }

    setOpen(1);    
    
    // restore the rtree header stored in first page
    memcpy((char*)&hdr, page+sizeof(rt_magic)+sizeof(PageID), sizeof(RTHdr));

#   ifdef  USE_MMAP
    struct  stat	fs;
    GAssert(fstat(file_handle, &fs) >=0);
    hdr.map_len = fs.st_size;
    hdr.map_region = mmap(NULL, hdr.map_len, PROT_READ|PROT_WRITE,
		      MAP_SHARED, file_handle, (off_t)0 );
    if(hdr.map_region == (caddr_t)-1 ) {
	msgWarn("can't mmap the rtree");
	::close(file_handle);
	return GFail;
    }
#   endif
    
    cout<<"RTStore::open: hdr="<<hdr<<endl;

    cleanBuf();
    
    return GOk;
}

Result 
RTStore::close(void)
{
    if(! isOpen() )
	msgFail("no index currently open");
    
    char*	page;

    // write back the rtree header
    read(0, page);
    //page = _buf;
    memcpy(page,  rt_magic, sizeof(rt_magic));
    memcpy(page+sizeof(rt_magic)+sizeof(PageID), (char*)&hdr, sizeof(RTHdr));
    write(0, page);

#   ifdef USE_MMAP
    munmap(hdr.map_region, hdr.map_len);
#   endif	

    ::close(file_handle);
    setOpen(0);
    return GOk;
}

Result
RTStore::read(PageID pg, char*& p)
{
#   ifdef USE_MMAP
    if (isOpen()) {
	p = (char*) (hdr.map_region + pg*RTPAGE_SZ);
	return GOk;
    }
    else
	 msgFail("index not open yet");
#   else    
    if (isOpen()) {
	::lseek(file_handle, pg * RTPAGE_SZ, SEEK_SET);
	if( ::read(file_handle, _buf, RTPAGE_SZ) != RTPAGE_SZ) {
	    msgWarn("can't read the page");
	    abort();
	}
	p = _buf;
	return GOk;
    }
    else
	msgFail("index not open yet");
#   endif
}

Result
RTStore::write(PageID pg, const char* p)
{
#   ifdef USE_MMAP
    //in most time it will be writing into same page;
    //so should be able to add some performance hack here.
    memcpy((char*)(hdr.map_region+pg*RTPAGE_SZ), p, RTPAGE_SZ);
    return GOk;
#   else
    if (isOpen()) {
	::lseek(file_handle, pg * RTPAGE_SZ, SEEK_SET);
	if( ::write(file_handle, p, RTPAGE_SZ) != RTPAGE_SZ)
	    msgFail("can't write the page");
	return GOk;
    }
    else
	msgFail("index not open yet");
#   endif
}

PageID
RTStore::allocPage ()
{
    PageID pid;
    char*  page;
    
    if( !isOpen()) {
	msgWarn("index not open yet");
	return 0;
    }
    // try to locate a free page in the store
    GAssert( read(0, page) == GOk);
    memcpy(&pid, page+sizeof(rt_magic), sizeof(PageID));    
    
    if(pid) {
	// reuse this page; but need to update the head of free list
	// cout<<"reusing a free page\n";
	read(pid, page);
	write(0, page);
    } else {
#   	ifdef  USE_MMAP
	//extend the file; has to do an unmap then re-mmap here.

	//extend 10 pages each time (just performance hack)
	static const int rt_extend_pages = 10;
	hdr.last_map_region = hdr.map_region, hdr.last_map_len = hdr.map_len;
	munmap(hdr.map_region, hdr.map_len);
	::lseek(file_handle, (rt_extend_pages*RTPAGE_SZ)-1, SEEK_END);
	::write(file_handle, "\0", 1);
	hdr.map_len += rt_extend_pages*RTPAGE_SZ;
	hdr.map_region = mmap(NULL, hdr.map_len, PROT_READ|PROT_WRITE,
			  MAP_SHARED, file_handle, (off_t)0);
	//link into free page list
	long mlen= hdr.last_map_len+RTPAGE_SZ; 

	for(int i=0; i<rt_extend_pages-1; i++) {
	    pid = mlen/RTPAGE_SZ;
	    deallocPage(pid);
	    mlen += RTPAGE_SZ;
	}
	//first extended page is for immediate use
	pid = hdr.last_map_len/RTPAGE_SZ;
	//cout<<"extended with "<<rt_extend_pages<<" pages\n";
#   	else
	// append a new page to the end of the file and return it
	pid = ::lseek(file_handle, 0, SEEK_END) / RTPAGE_SZ ;
	cleanBuf();
	::write(file_handle, _buf, RTPAGE_SZ);
#	endif
    }
    //cout<<"RTStore:: allocPage = "<<pid<<"\n";
    hdr.n_nodes ++;
    return pid;
}

Result
RTStore::deallocPage(PageID pid)
{
    char* 	page;
    PageID 	temp;
    
    if(!isOpen()) 
	msgFail("index not open yet");

#   ifdef  USE_MMAP    
    //get and save current head of free page list
    GAssert( read(0, page) == GOk);
    memcpy(&temp, page+sizeof(rt_magic), sizeof(PageID));
    //write the 'pid' as new head of free page list
    memcpy(page+sizeof(rt_magic), &pid, sizeof(PageID));
    GAssert( write(0, page) == GOk );
    //link to the old head: use _buf explictly for mmap'ed rtree
    memcpy(_buf+sizeof(rt_magic), &temp, sizeof(PageID));
    GAssert(write(pid, _buf) == GOk );
    
#   else
    //get and save current head of free page list
    GAssert( read(0, page) == GOk);
    memcpy(&temp, page+sizeof(rt_magic), sizeof(PageID));
    
    //write the 'pid' as new head of free page list
    memcpy(page+sizeof(rt_magic), &pid, sizeof(PageID));
    GAssert( write(0, page) == GOk );
    
    //link to the old head
    memcpy(page+sizeof(rt_magic), &temp, sizeof(PageID));
    GAssert( write(pid, page) == GOk );
#   endif
    
    return GOk;
}
