/****************************************************************************
 *   gbitarray.cpp : 
 *   		impl file for GFC bit array class.
 *
 *   Change Log :
 *		 Oct 28, 1998 Initial version,  Liujian Qian
 *
 *   $Id: gbitarray.cpp,v 1.2 1998/12/02 04:10:20 qian Exp $
 ***************************************************************************/
#include <strings.h>
#include "gbitarray.h"

// Note:
// inside each byte, the left-most (high-end) bit has the smallest
// bit index. 
//

GBitArray::GBitArray 		()
{
    data = 0;    n_bits = n_bytes = 0;
}

GBitArray::GBitArray 		(int size)
{
    n_bytes = (size+7)/8;
    n_bits = size;
    data = new unsigned char[n_bytes];
    GAssert(data);
}
    
GBitArray::GBitArray 		(const GBitArray& a)
{
    data = 0; n_bits = n_bytes = 0;
    resize(a.size());
    memcpy(data, a.data, n_bytes);
}

    
GBitArray::~GBitArray		()
{
    if(data) delete data;
}

int		GBitArray::size	() const
{
    return n_bits;
}

void	GBitArray::resize	(int new_size)
{
    if( new_size <=  (n_bytes<<3) ) {
	n_bits = new_size;
	return;
    } else {
	unsigned char* tmp = new unsigned char[(new_size+7)/8];
	GAssert(tmp);
	if(data) memcpy(tmp, data, n_bytes);
	n_bytes = (new_size+7)/8, n_bits = new_size;
	if(data) delete data; 	data = tmp ;
    }
}

void	GBitArray::fill 	(bool v, int size)
{
    if(size<0 || size>=n_bits) 
	memset(data, v?0xff:0, n_bytes);
    else {
	memset(data, v?0xff:0, size/8);
	unsigned char *p=data+size/8;
	unsigned char mask=0x80;
	for(int i=0; i<size%8; i++) {
	    if(v) *p |= mask;
	    else  *p &= ~mask;
	    mask >>= 1;
	}
    }
}

bool	GBitArray::test 	(int idx) const
{
    if(idx>=n_bits || idx<0) {
	msgWarn("GBitArray::test: invalid index.\n");
	return 0;
    }
    
    return *(data+ (idx>>3)) & (0x80 >> (idx & 7));
}

void	GBitArray::set  	(int idx)
{
    if(idx>=n_bits || idx<0) {
	msgWarn("GBitArray::set: invalid index.\n");
	return;
    }
    
    *(data+(idx>>3)) |= (0x80 >> (idx & 7));
}
	
void	GBitArray::clear   (int idx)
{
    if(idx>=n_bits || idx<0) {
	msgWarn("GBitArray::clear: invalid index.\n");
	return;
    }    
    *(data+(idx>>3)) &=  ~ (0x80 >> (idx & 7));
}

bool	GBitArray::toggle	(int idx)
{
    if(idx>=n_bits || idx<0) {
	msgWarn("GBitArray::toggle: invalid index.\n");
	return 0;
    }    
    return *(data+(idx>>3)) ^=  (0x80 >> (idx & 7));    
}

void	GBitArray::toggleAll()
{
    for(int idx=0; idx<n_bits; idx++) 
	*(data+(idx>>3)) ^=  (0x80 >> (idx & 7));
}

bool	GBitArray::operator[] (int idx) const
{
    if(idx>=n_bits || idx<0) {
	msgWarn("GBitArray::operator[]: invalid index.\n");
	return 0;
    }    
    return *(data+ (idx>>3)) & (0x80 >> (idx & 7));
}

GBitArray & GBitArray::operator= ( const GBitArray & a)
{
    resize(a.size());
    register unsigned char *data2 = a.data, *data1 = data;
    int n = n_bytes;
    while(--n >= 0) *data1++ = *data2++;
    
    return *this;
}    

GBitArray & GBitArray::operator&=( const GBitArray & a)
{
    if(n_bits == a.n_bits) {
	register unsigned char *data2 = a.data, *data1 = data;
	int n = n_bytes;
	while(--n >= 0) *data1++ &= *data2++;
    } else {
	msgWarn("GBitArray::operator&=: mismatched array size.\n");
    }
    return *this;
}
	
    
GBitArray & GBitArray::operator|=( const GBitArray & a)
{
    if(n_bits == a.n_bits) {
	register unsigned char *data2 = a.data, *data1 = data;
	int n = n_bytes;
	while(--n >= 0) *data1++ |= *data2++;
    } else {
	msgWarn("GBitArray::operator|=: mismatched array size.\n");
    }
    return *this;
}
    
GBitArray & GBitArray::operator^=( const GBitArray & a)
{
    if(n_bits == a.n_bits) {
	register unsigned char *data2 = a.data, *data1 = data;
	int n = n_bytes;
	while(--n >= 0) *data1++ ^= *data2++;
    } else {
	msgWarn("GBitArray::operator^=: mismatched array size.\n");
    }
    return *this;
}

GBitArray   GBitArray::operator~() const
{
    GBitArray a(n_bits);
    
    if(n_bits == a.n_bits) {
	register unsigned char *data2 = data, *data1 = a.data;
	int n = n_bytes;
	while(--n >= 0) *data1++ =~*data2++;
    } else {
	msgWarn("GBitArray::operator~=: mismatched array size.\n");
    }
    return a;
}


GBitArray operator&( const GBitArray &a1, const GBitArray &a2 )
{
    GBitArray a(a1.n_bits);
    if(a1.n_bits != a2.n_bits) 
	msgWarn("operator&(GBitArray&,GBitArray&): mismatched size.\n");
    int n = a1.n_bytes;
    register unsigned char *data=a.data, *data1=a1.data, *data2=a2.data;
    while(--n>=0) 
	*data++ = *data1++ & *data2++;
    return a;
}


GBitArray operator|( const GBitArray &a1, const GBitArray &a2 )
{
    GBitArray a(a1.n_bits);
    if(a1.n_bits != a2.n_bits) 
	msgWarn("operator&(GBitArray&,GBitArray&): mismatched size.\n");
    int n = a1.n_bytes;
    register unsigned char *data=a.data, *data1=a1.data, *data2=a2.data;
    while(--n>=0) 
	*data++ = *data1++ | *data2++;
    return a;
}
    
GBitArray operator^( const GBitArray &a1, const GBitArray &a2 )
{
    GBitArray a(a1.n_bits);
    if(a1.n_bits != a2.n_bits) 
	msgWarn("operator&(GBitArray&,GBitArray&): mismatched size.\n");
    int n = a1.n_bytes;
    register unsigned char *data=a.data, *data1=a1.data, *data2=a2.data;
    while(--n>=0) 
	*data++ = *data1++ ^ *data2++;
    return a;
}

ostream& operator<<(ostream& os, const GBitArray& a)
{
    for(int i=0; i<a.size(); i++){
	os<<a.test(i)<<" ";
	if( (i+1)%8==0 ) os<<"  ";
	if( (i+1)%32==0 ) os<<endl;
    }
    return os;
}

