/****************************************************************************
 *   gdbf.cc  : 
 *   	implementation of DBF file class.
 *
 *   Change Log :
 *		 May 1997 Initial version,  Liujian Qian
 *
 *   $Id: gdbf.cpp,v 1.9 1999/02/23 04:27:54 qian Exp $
 ***************************************************************************/
 
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#if defined(__unix__) || defined(unix)
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#endif

#include "gdbf.h"
#include "gshapefile.h"

GDBFile::GDBFile()
{
    init();
}

GDBFile::GDBFile(const char* name)
{
    init();
    
    openDBF(name);
}

GDBFile::~GDBFile()
{
    if(buffer) free(buffer);
    if(fields) free(fields);
    if(values) free(values);
}

void
GDBFile::init()
{
    n_fields = n_records = record_sz = buffer_sz = 0;
    buffer = 0;
    fd  = -1;
    fields = 0;
    values = 0;

    int i=1;
    if( *((unsigned char*)&i) == 1 )
	big_endian = false;
    else big_endian = true;
}

int
GDBFile::extendBuffer(int new_sz)
{
    if(new_sz <= buffer_sz) return 0;
    else {
	char* tmp=0;
	buffer_sz = Align(new_sz);
	tmp = (char*) malloc(buffer_sz);
	if(!tmp) {
	    cerr<<"out of memory"<<endl;
	    return -1;
	}
	
	if(buffer) {
	    free( buffer );
	}
	buffer = tmp;
	return 0;
    }
}

Result
GDBFile::openDBF(const char* name)
{
    char tmp[256];

    if(fd>0) close(fd);	//close previously open file
    
    if( strstr(name, ".dbf")==0 ) {
	strcpy(tmp, name);
	strcat(tmp,  ".dbf");
    }
    else strcpy(tmp, name);
    
    fd = open(tmp, O_RDONLY);
    if(fd<0) {
	msgFail("can't open dbf file ");
    }
    
    //read file header
    read(fd, &file_hdr, sizeof(GDBFileHdr));

    if(big_endian) {
	flipInt(&file_hdr.n_records);
	flipShort(&file_hdr.record_sz);
	flipShort(&file_hdr.header_sz);
    }
    
    n_records = file_hdr.n_records;
    record_sz = file_hdr.record_sz;
    n_fields = (file_hdr.header_sz - 1 - sizeof(GDBFileHdr) )/sizeof(GDBFieldDesc);
    
    extendBuffer(record_sz);

    if(fields) free(fields);
    if(values) free(values);
    
    fields = (GDBFieldDesc*) malloc((size_t)(sizeof(GDBFieldDesc)*n_fields));
    values = (GDBFValue*) malloc((size_t)(sizeof(GDBFValue)*n_fields));

    if(!fields || !values) {
	msgFail("out of memory!");
    }
    
    //read field descriptors
    read(fd, fields, sizeof(GDBFieldDesc)*n_fields);

    if(big_endian) {
	int i;
	for(i=0; i<n_fields; i++)
	    flipInt((unsigned char*)&fields[i].field_offset);
    }
    return  GOk;
}


GDBFieldDesc*
GDBFile::getFieldDesc(int n)
{
    if(n<0 || n>=n_fields) return NULL;
    return (GDBFieldDesc*) &fields[n];
}

char*
GDBFile:: getRecord (int n)
{
    int offset;
    
    offset =  sizeof(GDBFileHdr) + n_fields*sizeof(GDBFieldDesc) + 1 +
	n*record_sz;
    
    lseek(fd, offset, SEEK_SET);
    if(read(fd, buffer, record_sz) != record_sz) {
	cerr<<"error reading "<<n<<"th record in dbf file!"<<endl;
	return NULL;
    }
    
    return (buffer);
}

GDBFValue*		
GDBFile::recordDeform(char* rec)
{
    if(!rec) {
	cerr<<"invalid record pointer passed!"<<endl;
	return NULL;
    }
    if(rec[0] == DELETED) {
	cerr<<"this record has been deleted!"<<endl;
	return NULL;
    }
    int offset = 1;
    char tmp[256];
    
    for(int i=0; i<n_fields; i++) {
	if(fields[i].field_type == 'C') { // Character type 
	    strncpy(values[i].str_v, &rec[offset], fields[i].field_len);
	    values[i].str_v[fields[i].field_len] = '\0';
	}
	else if(fields[i].field_type == 'N') { //Number type
	    strncpy(tmp, &rec[offset], fields[i].field_len);
	    tmp[fields[i].field_len] = '\0';
	    if(fields[i].decimal_cnt>0) values[i].double_v = strtod(tmp, NULL);
	    else values[i].int_v = atoi(tmp);
	}
	offset += fields[i].field_len;
    }

    return values;
}

char*		
GDBFile::getFieldName(int ith)
{
    if(ith<0 || ith>=n_fields) {
	msgWarn("invalid field index !");
	return 0;
    }

    return fields[ith].name;
}

char
GDBFile::getFieldType(int ith)
{
    if(ith<0 || ith>=n_fields) {
	msgWarn("invalid field index !");
	return 0;
    }
    char t = fields[ith].field_type;
    if(t == 'N' && (fields[ith].decimal_cnt==0) ) t = 'I';
    return t;
}

char*		
GDBFile::getField(int ith, const char* rec)
{
    if(ith<0 || ith>=n_fields) {
	msgWarn("invalid field index !");
	return 0;
    }

    int offset = 1;
    for(int i=0; i<ith; i++) offset += fields[i].field_len;
    memset(field_buffer, 0, MAX_CHRS);

    strncpy(field_buffer, &rec[offset], fields[ith].field_len);
    field_buffer[fields[ith].field_len] = '\0';
    return field_buffer;
}


ostream&     operator<<(ostream& os, const GDBFile& dbf)
{
    os<<"\nDBF attributes: "<<endl;
    for(int i=0, j=0; i<dbf.n_fields; i++, j++) {
	os<<dbf.fields[i].name<<", ";
	if(j==5) { j=0; os<<"\n"; }
    }
    return os<<"\n";
}
