/****************************************************************************
 *   gshapefile.cc : 
 *   		implementation of shapefile class.
 *
 *   Change Log :
 *	
 *		 Jun 5th, fixed the misaligned double
 *		 pointers inside shape records.
 *
 *		 May 1997 Initial version,  Liujian Qian
 *
 *		 Jan 29, 1998, Major restructuring.
 *
 *   $Id: gshapefile.cpp,v 1.19 1999/02/17 02:06:30 qian Exp $
 ***************************************************************************/

#include <iostream.h>
#include <strstream.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef __unix__
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#endif
 
#include "gshapefile.h"
#include "gdatapipe.h"
#include "gadt_rect.h"
#include "gadt_point.h"

GShapeFile::GShapeFile()
{
    srec = 0;
    init();
}

GShapeFile::GShapeFile(const char* name)
{
    srec = 0;
    init();
    
    openShapeFile(name);
}

GShapeFile::~GShapeFile()
{
    if(main_mmap) delete main_mmap;
    if(idx_mmap) delete idx_mmap;
    if(dbf_mmap) delete dbf_mmap;
    if(srec) delete srec;
}

    
ostream&	operator<<(ostream& os, const GShapeFile& shp)
{
    char t[20];
    memset(t, 0, 20);
    switch (shp.type) {
    case 1:	strcpy(t, "Point"); 	break;
    case 3:	strcpy(t, "Arc"); 	break;
    case 5:	strcpy(t, "Polygon"); 	break;
    case 8:	strcpy(t, "Multipoint");break;
    default:	strcpy(t, "Unknown");
    }
    double xl,yl,xh,yh;
    shp.bndRect(xl,yl,xh,yh);
    
    os<<"type="<<t<<", #records="<<shp.n_records<<", bbox=";
    os<<xl<<","<<yl<<","<<xh<<","<<yh<<endl;
    return os;
}

void
GShapeFile::init()
{
    n_records = type = 0;
    rec_buf.recycle();
    name_prefix[0] = '\0';
    main_mmap = idx_mmap = dbf_mmap = 0;
}

//
// retrieve the bounding rectangle for the WHOLE shape file
//
int
GShapeFile::bndRect(double& xmin, double& ymin, double& xmax, double& ymax) const
{
    memcpy(&xmin, main_hdr.Xmin, sizeof(double));
    memcpy(&ymin, main_hdr.Ymin, sizeof(double));
    memcpy(&xmax, main_hdr.Xmax, sizeof(double));
    memcpy(&ymax, main_hdr.Ymax, sizeof(double));
    return 0;
}

int
GShapeFile::openShapeFile(const char* name)
{
    strcpy(name_prefix, name);
    
    //close previously mapped files
    if(main_mmap) delete main_mmap;    if(idx_mmap) delete idx_mmap;
    if(dbf_mmap) delete dbf_mmap;
    
    //open the .shp and .shx files
    char tmp[256]; 
    strcpy(tmp, name_prefix);
    if(!strstr(tmp, ".shp"))	
	strcat(tmp, ".shp");
    main_mmap = new GMMap(tmp);
    //memory-map the main file
    //this can be optimized so as not to map all of it.
    GAssert(main_mmap->mapAll("r"));
    
    strcpy(tmp, name_prefix);
    strcat(tmp, ".shx");
    idx_mmap = new GMMap(tmp);
    GAssert(idx_mmap->mapAll("r"));

    //read the main/index file headers
    memcpy((char*)&main_hdr, main_mmap->ptr(), AV_HEADER_SIZE);
    memcpy((char*)&idx_hdr,  idx_mmap->ptr(),  AV_HEADER_SIZE);

    flipHeaders();

    //setup n_records information
    n_records = (idx_hdr.file_length * 2 - AV_HEADER_SIZE)/sizeof(IndexRecord);
    type      = main_hdr.shape_type;

    allocSpatialRecord();
    return 0;
}

Result
GShapeFile::getMBR(int nth, GRect& rect)
{
    char* buf;
    double xl,yl,xh,yh;
    
    //fetch the physical record into buffer
    if( (buf=recordCopy(nth)) == NULL) {
	cerr<<"error fetching shape record!"<<endl;
	return GFail;
    }
      
    if( type == av_Point ) {
	if(GEndian::big) {
	    flipDouble( (double*)(((GShpPoint*)buf)->x));
	    flipDouble( (double*)(((GShpPoint*)buf)->y));
	}
	memcpy(&xl, ((GShpPoint*)buf)->x, sizeof(double));
	memcpy(&yl, ((GShpPoint*)buf)->y, sizeof(double));
	xh = xl, yh = yl;
    }
    else {	// other types are similar
	if(GEndian::big) 
	    flipCommon(buf);
	copyMBR(xl,yl,xh,yh,buf);
    }
    rect.set(xl,yl,xh,yh);    
    return GOk;
}

Result
GShapeFile::getShapeRec (int nth,      //the nth shape record
			int& n_parts, //# of parts in the record
			int& n_verts, //# of vertices (of all parts)
			int*& parts,   //parts index (into verts)
			double*& verts,//all verticies (x1,y1,...)
			GRect& r) 	//bounding rectangle
{
    char* buf;
    int i;
    double* p, *p_aligned;
    
    memcpy(&r.xlow, main_hdr.Xmax, sizeof(Coord));
    memcpy(&r.ylow, main_hdr.Ymax, sizeof(Coord));
    memcpy(&r.xhi,  main_hdr.Xmin, sizeof(Coord));
    memcpy(&r.yhi,  main_hdr.Ymin, sizeof(Coord));

    //fetch the physical record into buffer
    if( (buf=recordCopy(nth)) == NULL) {
	msgFail("error fetching shape record!");
    }
    
    n_parts = 1, n_verts = 0;
    parts =  0, verts = 0;
    
    switch(type) {
    case av_Point:		//point
	p = (double*)(buf+4);
	p_aligned = (double*) GAlign((unsigned)p);
	if(p_aligned !=p) {
	    memmove((char*)p_aligned, (char*)p, 2*PolyDim*sizeof(double));
	    p = p_aligned;
	}
	if(GEndian::big) {
	    flipDouble(p);
	    flipDouble(p+1);
	}
	n_verts = 1, verts = p;
	return GOk;
    
    case  av_MultiPoint:	// MultiPoint
	if(GEndian::big) {
	    flipCommon(buf);
	    flipInt( (int*) ((unsigned char*)buf+36));
	}
	memcpy(&n_verts, buf+36, sizeof(int));
	p = verts = (double*)(buf+40);
	if(GEndian::big) {
	    for(i=0; i<n_verts; i++){
		flipDouble(& p[i*2]);
		flipDouble(& p[i*2+1]);
	    }
	}
	return GOk;
	
    case av_Arc:
    case av_Polygon:		// ARC || Polygon
	if(GEndian::big) {
	    flipCommon(buf);
	    flipInt((int*)((unsigned char*)buf+36)); //n_parts
	    flipInt((int*)((unsigned char*)buf+40)); //n_points
	}
	memcpy(&n_parts, buf+36, sizeof(int));
	memcpy(&n_verts, buf+40, sizeof(int));
	parts = (int*)(buf+44);
	p = (double*) (buf + 44 + 4*n_parts);
	p_aligned = (double*) GAlign((unsigned)p);
	if(p_aligned !=p) {
	    memmove((char*)p_aligned, (char*)p, n_verts*2*sizeof(double));
	    p = p_aligned;
	}

	verts = p;
	if(GEndian::big) {
	    for(i=0; i<n_parts; i++) 
		flipInt(&parts[i]);

	    for(i=0; i<n_verts; i++) {
		flipDouble(& p[i*2]);
		flipDouble(& p[i*2+1]);
	    }
	}
	///!!! some shapefile seems to have a funky pair of (xxx.xx, 0) 
	///!!! coordinate as the last one for a multi-part polygon/arc record.
	///!!! really wierd !!!

	//generate mbr for this record.	
	for(i=0; i<n_verts; i++) {
	    if(r.xlow > p[i*2]) r.xlow = p[i*2];
	    if(r.ylow > p[i*2+1]) r.ylow = p[i*2+1];
	    if(r.xhi  < p[i*2])   r.xhi  = p[i*2];
	    if(r.yhi  < p[i*2+1]) r.yhi  = p[i*2+1];
	}
	return GOk;

    default:		//invalid type
	return GFail;
    }
}

GSpatial* 
GShapeFile::getShapeRec(int nth) 
{
    if(!srec) return 0;
    switch(type) {
    case av_Point:
	GAssert(srec->type()==_Point);
	getShapeRec(nth, (GPoint*)srec); 	break;
    case av_Arc:
	GAssert(srec->type() == _Arc);
	getShapeRec(nth, (GArc*)srec);   	break;
    case av_Polygon:
	GAssert(srec->type() == _Polygon);
	getShapeRec(nth, (GPolygon*)srec);	break;
    default:
	msgWarn("unsupported shape type!");	
	return 0;
    }
    return srec;
}



//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// protected methods

void 
GShapeFile::flipCommon(char* rec)
{
    flipDouble( (double*)(((GShpCommon*)rec)->Xmin));
    flipDouble( (double*)(((GShpCommon*)rec)->Ymin));
    flipDouble( (double*)(((GShpCommon*)rec)->Xmax));
    flipDouble( (double*)(((GShpCommon*)rec)->Ymax));
}

void
GShapeFile::copyMBR(double& xl, double& yl, double& xh, double& yh, 
		   char* rec)
{
    memcpy((unsigned char*)&xl,((GShpCommon*)rec)->Xmin, sizeof(double));
    memcpy((unsigned char*)&yl,((GShpCommon*)rec)->Ymin, sizeof(double));
    memcpy((unsigned char*)&xh,((GShpCommon*)rec)->Xmax, sizeof(double));
    memcpy((unsigned char*)&yh,((GShpCommon*)rec)->Ymax, sizeof(double));    
} 

char*
GShapeFile::recordCopy(int n)
{
    int byte_off, byte_len;

    byte_off = mainRecordOffset(n, byte_len);
    
    rec_buf.extendTo(byte_len+17); //extra space for Solaris 8-byte alignment!!!
    rec_buf.recycle();
    
    memcpy(rec_buf.ptr(), (char*)(main_mmap->ptr())+byte_off, byte_len);
    //we are not intrested in the generic record header attached to
    //each shpae record. (ArcView silliness here).
    return rec_buf.ptr() + sizeof(RecordHdr);
}

int
GShapeFile::advise(int ad) 
{
    if(main_mmap) main_mmap->advise(ad);
    if(idx_mmap) idx_mmap->advise(ad);
    if(dbf_mmap) dbf_mmap->advise(ad);
    return 0;
}


char* GShapeFile::makePath(const char* , char* )
{
    return 0;
}

void  GShapeFile::flipHeaders()
{
    //flip byte-order if necessary
    if(!GEndian::big) { //these are in Big-endian byte order!!
        flipInt( &(main_hdr.file_code) );
        flipInt( &(idx_hdr.file_code) );
        flipInt( &(main_hdr.file_length) );
        flipInt( &(idx_hdr.file_length) );
    }
    if(GEndian::big) { //these are in Little-endian byte order!!
        flipDouble( main_hdr.Xmin );
        flipDouble( main_hdr.Ymin );
        flipDouble( main_hdr.Xmax );
        flipDouble( main_hdr.Ymax );
        flipDouble( idx_hdr.Xmin );
        flipDouble( idx_hdr.Ymin );
        flipDouble( idx_hdr.Xmax );
        flipDouble( idx_hdr.Ymax );

        flipInt( &(main_hdr.version) );
        flipInt( &(idx_hdr.version) );
        flipInt( &(main_hdr.shape_type) );
        flipInt( &(idx_hdr.shape_type) );
    }
}

void  GShapeFile::allocSpatialRecord()
{
    if(srec) delete srec; srec = 0;
    switch(type) {
    case av_Point:	srec = new GPoint; 	break;
    case av_Arc:	srec = new GArc;   	break;
    case av_Polygon:	srec = new GPolygon; 	break;
    default:  srec = 0;
    }
}

long
GShapeFile::mainRecordOffset(int nth, int& len)
{
    IndexRecord idx;
    
    memcpy(&idx, 
	   (char*)idx_mmap->ptr()+AV_IDX_OFFSET(nth), 
	   sizeof(IndexRecord));
    
    if(!GEndian::big) {
	flipInt( &(idx.offset) );
	flipInt( &(idx.content_len) );
    }
    len = idx.content_len*2;  //convert to byte unit
    return idx.offset*2;  //convert to byte offset.
}


Result 
GShapeFile::getShapeRec (int nth, GPoint* p)
{
    int n_verts, n_parts, *parts;
    double *verts;   
    GRect   mbr;    

    GAssert(getShapeRec(nth, n_parts, n_verts, parts, verts, mbr) == GOk);
    p->set(verts[0], verts[1]);
    return GOk;
}

Result
GShapeFile::getShapeRec (int nth, GArc* a)
{
    int n_verts, n_parts, *parts, count=0;
    double *verts;   
    GRect   mbr;    
    
    GAssert(getShapeRec(nth, n_parts, n_verts, parts, verts, mbr) == GOk);        

    //convert the array of parts index into an array of parts size
    for(int i=0; i<n_parts; i++) {
	if(i<n_parts-1) count = parts[i+1]-parts[i];
	else count = n_verts-parts[i];
	parts_a[i] = count;
    }
    
    a->set(n_parts, n_verts, parts_a, verts);
    a->setMbr(mbr);
    return GOk;
}

Result
GShapeFile::getShapeRec (int nth, GPolygon* pg)
{
    int n_verts, n_parts, *parts;
    double *verts;   
    GRect   mbr;    
    
    GAssert(getShapeRec(nth, n_parts, n_verts, parts, verts, mbr) == GOk);        

    //convert the array of parts index into an array of parts size
    for(int i=0, count=0; i<n_parts; i++) {
	if(i<n_parts-1) count = parts[i+1]-parts[i];
	else count = n_verts-parts[i];
	parts[i] = count;
    }

    pg->set(n_parts, n_verts, parts, verts);
    pg->setMbr(mbr);
    return GOk;
}
